/***************************************************************************
 * atom.c
 * atom.c,v 1.2 1995/03/17 14:17:12 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme symbols implementation.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <os_config.h>
#include <os_interp.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_htable.h>
#include <os_cons.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_port.h>
#include <os_arit.h>

OsClass OsAtomClass = (OsClass) 0 ;

OsObject OsA_Nil           = (OsObject) 0 ;
OsObject OsA_True          = (OsObject) 0 ;
OsObject OsA_False         = (OsObject) 0 ;
OsObject OsA_Unbound       = (OsObject) 0 ;
OsObject OsA_Quote         = (OsObject) 0 ;
OsObject OsA_Quasiquote    = (OsObject) 0 ;
OsObject OsA_Unquote       = (OsObject) 0 ;
OsObject OsA_UnquoteSplice = (OsObject) 0 ;
OsObject OsA_Lambda        = (OsObject) 0 ;

static OsString os_gensym_prefix = (OsString) 0 ;

OsAtom OsAtomMake (OsInterp interp, char *name, OsObject value) {
    OsAtom atom ;

    if ( ! (atom = (OsAtom) OsHashtableGet (interp->name_space, name)) ) {
        atom = (OsAtom) OsObjectMake (OsAtomClass, 0) ;
        atom->pname = OsStringMake (name) ;
        atom->vcell = value ;
        OsHashtablePut (interp->name_space, atom->pname->string, atom) ;
    } else {
        if ( atom->vcell == OsA_Unbound )
            atom->vcell = value ;
    }
    return (atom) ;
}

OsAtom OsAtomMakeWithString (OsInterp interp, OsString pname, OsObject value) {
    OsAtom atom ;
    
    if (!(atom = (OsAtom) OsHashtableGet(interp->name_space, pname->string))) {
        atom = (OsAtom) OsObjectMake (OsAtomClass, 0) ;
        atom->pname = pname ;
        atom->vcell = value ;
        OsHashtablePut (interp->name_space, pname->string, atom) ;
    } else {
        if ( atom->vcell == OsA_Unbound )
            atom->vcell = value ;
    }
    return (atom) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (symbol? OBJECT)
 * Returns #t if OBJECT is a symbol, otherwise returns #f.
 *
 *   (symbol? 'foo)              ==>  #t
 *   (symbol? (car '(a b)))      ==>  #t
 *   (symbol? "bar")             ==>  #f
 *   (symbol? 'nil)              ==>  #t
 *   (symbol? '())               ==>  #f
 *   (symbol? #f)                ==>  #f
 */

static OsObject OsSubrSymbolP (OsObject atom) {
    /* nil, #t and #f are not symbols */
    if ( OsIsAtom (atom) ) {
        if ((atom != OsA_Nil) && (atom != OsA_True) && (atom != OsA_False)) {
            return (OsA_True) ;
        }
    }
    return (OsA_False) ;
}

/* AUTODOC
 * (symbol->string SYMBOL)
 * Returns the name of SYMBOL as a string.
 * OScheme makes a difference betweenm lower case and upper case, so the 
 * returned string might contain both of them, depending on how the symbol
 * was created.
 */

static OsObject OsSubrSymbolToString (OsAtom atom) {
    /* To make symbol's name constant, we use copy of strings for pname */
    OsMustBeAtom (atom) ;
    return ((OsObject) OsStringMakeFromBytes (OsStringLength (atom->pname)
                                              , OsStringBytes (atom->pname))) ;
}

/* AUTODOC
 * (string->symbol STRING)
 * Returns the symbol whose name is string. 
 * This procedure can create symbols with names containing special
 * characters or letters in the non-standard case, but it is usually a
 * bad idea to create such symbols because in some implementations of
 * Scheme they cannot be read as themselves.  */

static OsObject OsSubrStringToSymbol (OsString str) {
    OsString pname ;
    OsShouldBeString (str) ;
    pname = OsStringMakeFromBytes (OsStringLength (str)
                                   , OsStringBytes (str)) ;
    return ((OsObject) OsAtomMakeWithString (OsCurrentInterp()
                                             , pname
                                             , OsA_Unbound)) ;
}

/* AUTODOC
 * (gensym [PREFIX])
 * Returns a new symbol, starting with the optional PREFIX.
 */

static OsObject OsSubrGensym (unsigned int argc, OsObject *argv) {
    static int counter = 1 ;
    char       num[8] ; 
    OsString   pref = os_gensym_prefix ;
    
    switch (argc) {
      case 1:
        OsMustBeString (*argv) ;
        pref = (OsString) *argv ;
        /* no break */
      case 0:
        sprintf (num, "%04d", counter++) ;
        return ((OsObject) OsAtomMakeWithString (OsCurrentInterp()
                                                 , OsStringCAppend (pref, num)
                                                 , OsA_Unbound)) ;
      default:
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (-1)
                               , OsIntegerMake (argc)) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (unbound)
 * Returns the object representing UNBOUND.
 */

static OsObject OsSubrUnbound () {
    return (OsA_Unbound) ;
}

/* AUTODOC
 * (boolean? OBJECT)
 * Returns #t if OBJECT if #t of #f, #f otherwise.
 */

static OsObject OsSubrBooleanP (OsObject obj) {
    return (((obj == OsA_True) || (obj == OsA_False)) ? OsA_True : OsA_False) ;

}

/* AUTODOC
 * (bound? SYMBOL)
 * Is the global SYMBOL bound to some value ?
 * This primitive tests whether the given global SYMBOL is bound or not. It
 * returns #t if it is, #f otherwise.
 */

static OsObject OsSubrBoundP (atom)
    OsAtom atom ;
{
    OsShouldBeAtom (atom) ;
    return ((atom->vcell == OsA_Unbound) ? OsA_False : OsA_True) ;
}

/*
 * Messages impl for atoms:
 */

static OsObject os_atom_display (OsAtom atom
                                 , int sel
                                 , OsOutputPort port) {
    char tmp[OS_DISPLAY_BUFSIZ] ;
    sprintf (tmp, "%s", atom->pname->string) ;
    OsOutputPortWriteCString (port, tmp) ;
    return ((OsObject) atom) ;
}

static OsObject os_atom_equal (OsAtom atom, int sel, OsAtom other) {
    return ((other == atom) ? OsA_True : OsA_False) ;
}

void OsAtomInitialize (OsInterp i, int *pargc, char **argv) {
    OsAtomClass = OsClassDeclare ("Atom", sizeof (OsAtomRec)) ;
    OsDefineMsg (OsAtomClass, OS_SEL_DISPLAY, os_atom_display) ;
    OsDefineMsg (OsAtomClass, OS_SEL_WRITE, os_atom_display) ;
    OsDefineMsg (OsAtomClass, OS_SEL_EQUAL, os_atom_equal) ;
    
    OsA_Unbound       = (OsObject) OsAtomMake (i,"*unbound*",OsA_Unbound) ;
    OsA_Nil           = (OsObject) OsAtomMake (i,"nil",OsA_Unbound) ;
    OsA_True          = (OsObject) OsAtomMake (i,"#t",OsA_Unbound) ;
    OsA_False         = (OsObject) OsAtomMake (i,"#f",OsA_Unbound) ;
    OsA_Quote         = (OsObject) OsAtomMake (i,"quote",OsA_Unbound) ;
    OsA_Quasiquote    = (OsObject) OsAtomMake (i,"quasiquote",OsA_Unbound) ;
    OsA_Unquote       = (OsObject) OsAtomMake (i,"unquote",OsA_Unbound) ;
    OsA_UnquoteSplice = (OsObject) OsAtomMake (i,"unquote-splice",OsA_Unbound);
    OsA_Lambda        = (OsObject) OsAtomMake (i,"lambda",OsA_Unbound) ;

    ((OsAtom) OsA_False)->vcell   = OsA_False ;
    ((OsAtom) OsA_True)->vcell    = OsA_True ;
    ((OsAtom) OsA_Nil)->vcell     = OsA_Nil ;
    ((OsAtom) OsA_Unbound)->vcell = OsA_Unbound ;

    os_gensym_prefix = OsStringMake ("gensym-") ;
    
    return ;
}

void OsAtomSubrInitialize (OsInterp interp) {
    OsSubrMake (interp, "symbol?", OsSubrSymbolP, 1) ;
    OsSubrMake (interp, "symbol->string", OsSubrSymbolToString, 1) ;
    OsSubrMake (interp, "string->symbol", OsSubrStringToSymbol, 1) ;
    OsSubrMake (interp, "gensym", OsSubrGensym, OS_NARY) ;
    OsSubrMake (interp, "unbound", OsSubrUnbound, 0) ;
    OsSubrMake (interp, "boolean?", OsSubrBooleanP, 1) ;
    OsSubrMake (interp, "bound?", OsSubrBoundP, 1) ;
    return ;
}

    
