/***************************************************************************
 * gc.c
 * gc.c,v 1.1.1.1 1995/02/27 15:15:49 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * Interface with the garbage collector. 
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <os_config.h>
#include <os_interp.h>
#include <gc_private.h>
#include <os_gc.h>
#include <os_obj.h>
#include <os_string.h>
#include <os_atom.h>
#include <os_vector.h>
#include <os_error.h>
#include <os_cons.h>
#include <os_arit.h>
#include <os_cblock.h>
#include <os_subr.h>
#include <os_misc.h>

static OsError ose_bad_malloc = (OsError) 0 ;

int OsGCExpandHeap (count)
    int count ;
{
    return (GC_expand_hp (count)) ;
}

#ifndef OS_MACRO_GC_CALL

void *OsGCMalloc (size)
    unsigned int size ;
{
    void *p = GC_MALLOC (size) ;

    if ( p )
	return (p) ;
    (void) OsErrorTrigger (1, ose_bad_malloc
                           , OsStringMake ("OsGCMalloc")
                           , OsIntegerMake (size)) ;
}

void *OsGCRealloc (p, size)
    void *p ;
    unsigned int size ;
{
    void *r = (void *) GC_REALLOC (p, size) ;

    if ( r )
	return (r) ;
    (void) OsErrorTrigger (1, ose_bad_malloc
                           , OsStringMake ("OsGCRealloc")
                           , OsIntegerMake (size)) ;
}

void OsGCFree (p)
    void *p ;
{
    GC_FREE (p) ;
}

void *OsGCMallocAtomic (size)
    unsigned int size ;
{
    void *p = (void *) GC_MALLOC_ATOMIC (size) ;

    if ( p )
	return (p) ;
    (void) OsErrorTrigger (1, ose_bad_malloc
                           , OsStringMake ("OsGCMallocAtomic")
                           , OsIntegerMake (size)) ;
}

void OsGCFreeAtomic (p)
    void *p ;
{
    GC_FREE (p) ;
    return ;
}

void *OsMalloc (size)
    unsigned int size ;
{
    void *p = (void *) malloc (size) ;
    if ( p )
	return (p) ;
    (void) OsErrorTrigger (1, ose_bad_malloc
                           , OsStringMake ("OsMalloc")
                           , OsIntegerMake (size)) ;
}

void OsFree (p)
    void *p ;
{
    free (p) ;
    return ;
}

void OsGCRegisterFinalizer (obj, cfinalizer)
    void *obj ;
    void (*cfinalizer)() ;
{
    GC_register_finalizer (obj, cfinalizer, 0, 0, 0) ;
    return ;
}
    
#endif /* OS_MACRO_GC_CALL */

char *OsGCStrdup (str)
    char *str ;
{
    int len = strlen (str) ;
    char *p = (char *) OsGCMallocAtomic (len+1) ;
    bcopy (str, p, len+1) ;
    return (p) ;
}

/*
 * Subr allowing to control gc:
 */

/* AUTODOC
 * (gc:disable)
 * Disable the garbage collector.
 */

static OsObject OsSubrGCDisable () {
    GC_dont_gc = 1 ;
    return (OsA_False) ;
}

/* AUTODOC
 * (gc:enable)
 * Enables the garbage collector.
 */

static OsObject OsSubrGCEnable () {
    GC_dont_gc = 0 ;
    return (OsA_True) ;
}

/* AUTODOC
 * (gc:collect)
 * Enforces a garbage collection.
 */

static OsObject OsSubrGCCollect () {
    extern signed_word GC_mem_found ;
    GC_gcollect () ;
    return ((OsObject) OsIntegerMake (GC_mem_found)) ;
}

/* AUTODOC
 * (gc:expand-heap SIZE)
 * Set new heap size to SIZE.
 * SIZE should be a positive integer. If the heaap was successfully resized,
 * this primitive returns #t, otherwise, it returns #f.
 */

static OsObject OsSubrGCExpandHeap (OsInteger count) {
    OsMustBeInteger (count) ;
    if ( (count->ival > 0) && (OsGCExpandHeap (count->ival)) )
        return (OsA_True) ;
    return (OsA_False) ;
}

/* AUTODOC
 * (gc:stat)
 * Garbage ciollector usage statistics.
 * This primitive returns a vector of four items:
 * #(NO ATOMIC_ALLOCATED COMPOSITE_ALLOCATED HEAP_SIZE)
 * Where:
 *     NO: is the current collection number,
 *     ATOMIC_ALLOCATED: is the number of pointer free bytes allocated,
 *     COMPOSITE_ALLOCATED: number of non pointer free bytes allocated,
 *     HEAP_SIZE: is the current heap size.
 */

static OsObject OsSubrGCStat () {
    OsVector vstat = OsVectorMake (4) ;

/* Stick to 80 columns */    
#define OS_WTOB(s)	OsIntegerMake((s) * sizeof (void *))

    (void) OsVectorFastAppend (vstat, (OsObject) OsIntegerMake (GC_gc_no)) ;
    (void) OsVectorFastAppend (vstat
                               , (OsObject) OS_WTOB (GC_atomic_in_use)) ;
    (void) OsVectorFastAppend (vstat
                               , (OsObject) OS_WTOB (GC_composite_in_use)) ;
    (void) OsVectorFastAppend (vstat
                               , (OsObject) OsIntegerMake(GC_get_heap_size()));
    return ((OsObject) vstat) ;
}

/*
 * Subr initialization:
 */

void OsGCInitialize (OsInterp interp, int *pargc, char **argv) {
    /* WARNING: interp not initialized yet ! */
    int          heapsize = OS_DEFAULT_HEAPSIZE ;
    unsigned int argc = 1 ;
    while ( argc < *pargc ) {
        if ( ! strcmp ("-gcfsd", argv[argc]) ) {
            int gcfsd ;
            OsArgvUsed (pargc, argc, argv) ;
            if ( (gcfsd = atoi (argv[argc])) > 0 ) {
                GC_free_space_divisor = gcfsd ;
            } else {
                fprintf (stderr, "invalid free space divisor %d\n", gcfsd) ;
            }
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-hs", argv[argc]) ) {
            int hpsz ;
            OsArgvUsed (pargc, argc, argv) ;
            if ( (hpsz = atoi (argv[argc])) <= 0 ) {
                fprintf (stderr, "invalid heap size %d\n", heapsize) ;
            } else {
                heapsize = hpsz ;
            }
            OsArgvUsed (pargc, argc, argv) ;
        } else {
            argc++ ;
        }
    }
    OsGCExpandHeap (heapsize) ;
    return ;
}

void OsSubrGCInitialize (OsInterp interp) {
    ose_bad_malloc = OsErrorMake (interp
                                  , "error:bad-malloc"
                                  , "request %1 bytes in %0 failed."
                                  , 2, OSE_FATAL) ;
    OsSubrMake (interp, "gc:collect", OsSubrGCCollect, 0) ;
    OsSubrMake (interp, "gc:stat", OsSubrGCStat, 0) ;
    OsSecSubrMake (interp, "gc:expand-heap", OsSubrGCExpandHeap
                   , 1, OS_SECURITY_LEVEL_B) ;
    OsSecSubrMake (interp, "gc:disable", OsSubrGCDisable
                   , 0, OS_SECURITY_LEVEL_A) ;
    OsSecSubrMake (interp, "gc:enable", OsSubrGCEnable
                   , 0, OS_SECURITY_LEVEL_A) ;
    return ;
}
