/***************************************************************************
 * interp.c
 * interp.c,v 1.4 1995/03/17 14:22:26 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

OsClass OsInterpClass = (OsClass) 0 ;

static OsInterp _theInterp = (OsInterp) 0 ;
static int      default_stack_size = OS_VM_STACKWSIZE ;

#ifndef OS_MAX_ULONG
#    define OS_MAX_ULONG (~((unsigned long) 0))
#endif

#ifdef OS_SECURE
static unsigned long default_read_bytes    = OS_MAX_ULONG ;
static unsigned long default_written_bytes = OS_MAX_ULONG ;
static unsigned long default_clicks        = OS_MAX_ULONG ;
static int           default_security      = OS_SECURITY_LEVEL_F ;
#endif /* OS_SECURE */

static OsError ose_read     = (OsError) 0 ;
static OsError ose_written  = (OsError) 0 ;
static OsError ose_clicks   = (OsError) 0 ;
static OsError ose_stack    = (OsError) 0 ;
static OsError ose_security = (OsError) 0 ;

/*
 * Create a new (empty) interpretor:
 */

OsInterp OsInterpMake () {
    OsInterp interp    = (OsInterp) OsObjectMake (OsInterpClass, 0) ;
    /* Allocate both stack, and debug stack
     * Debug stack allocation could be delayed until needed... */
    interp->stack      = OsGCMalloc (sizeof (OsObject) * default_stack_size) ;
    interp->stacke     = interp->stack + default_stack_size ;
    interp->sp         = interp->stack ;
    interp->dbg_stack  = OsGCMalloc (sizeof (OsObject) * default_stack_size) ;
    interp->dbg_sp     = interp->dbg_stack ;
    /* Setup the name and macro spaces: */
    interp->name_space = OsHashtableMake (127
                                          , (OsHashFct) OsHashCString
                                          , (OsCmpFct) OsCmpCString) ;
    interp->macro_space = OsHashtableMake (0
                                           , (OsHashFct) OsHashPointer
                                           , (OsCmpFct) OsCmpPointer) ;
    /* load path, ect: are inited by path.c (see comment above) */
    interp->path_list  = (void *) 0 ;
    interp->ext_list   = (void *) 0 ;
    interp->prov_list  = (void *) 0 ;
    interp->hooked_ext = (void *) 0 ;
    /* Error handler defaults to #f
    * As this function first get called within OsInitialize, *before* atoms
    * are initialized, we have to wait until OsInterpSubrInitialize() is
    * called to really get #f in the field */
    interp->error_handler  = OsA_False ;
    /* Setup limits, needed only under OS_SECURE */
#ifdef OS_SECURE
    interp->read_bytes     = 0 ;
    interp->lread_bytes    = default_read_bytes ;
    interp->written_bytes  = 0 ;
    interp->lwritten_bytes = default_written_bytes ;
    interp->clicks         = 0 ;
    interp->lclicks        = default_clicks ;
    interp->security       = default_security ;
#endif /* OS_SECURE */
    return (_theInterp = interp) ;
}

OsInterp OsInterpClone (OsInterp interp
                        , int stack_size
                        , int security
                        , unsigned long lread_bytes
                        , unsigned long lwritten_bytes
                        , unsigned long lclicks) {
    OsInterp clone = (OsInterp) OsObjectMake (OsInterpClass, 0) ;
    /* Allocate both stack, and debug stack
     * Debug stack allocation could be delayed until needed... */
    clone->stack      = OsGCMalloc (sizeof (OsObject) * stack_size) ;
    clone->stacke     = clone->stack + stack_size ;
    clone->sp         = clone->stack ;
    clone->dbg_stack  = OsGCMalloc (sizeof (OsObject) * stack_size) ;
    clone->dbg_sp     = clone->dbg_stack ;
    /* Setup the name and macro spaces: here, we duplicate the name space
     * not the symbols... */
    clone->name_space  = OsHashtableCopy (interp->name_space) ;
    clone->macro_space = OsHashtableCopy (interp->macro_space) ;
    /* load path, etc get copied from source interp */
    clone->path_list   = OsListCopy (interp->path_list) ;
    clone->ext_list    = OsListCopy (interp->ext_list) ;
    clone->prov_list   = OsListCopy (interp->prov_list) ;
    clone->hooked_ext  = OsHashtableCopy (interp->hooked_ext) ;
    /* Error handler defaults to #f */
    interp->error_handler = OsA_False ;
    /* Setup limits, needed only under OS_SECURE */
#ifdef OS_SECURE
    clone->read_bytes     = 0 ;
    clone->lread_bytes    = lread_bytes ;
    clone->written_bytes  = 0 ;
    clone->lwritten_bytes = lwritten_bytes ;
    clone->clicks         = 0 ;
    clone->lclicks        = lclicks ;
    clone->security       = security ;
#endif
    return (clone) ;
}

/*
 * OsCurrentInterp: get current interpreter (should be macro)
 * OsInterpSwitch:  switch to another interpreter
 */

OsInterp OsCurrentInterp () {
    return (_theInterp) ;
}

OsInterp OsInterpSwitch (OsInterp interp) {
    OsInterp old = _theInterp ;
    _theInterp = interp ;
    return (old) ;
}

/* As security checks are only performed in secure mode, those functions are
 * reached *only* if the oscheme has been compiled with OS_SECURE flag set. */

unsigned long OsInterpReadBytesExceeded (OsInterp interp) {
    (void) OsErrorTrigger (1, ose_read) ;
}

unsigned long OsInterpWrittenBytesExceeded (OsInterp interp) {
    (void) OsErrorTrigger (1, ose_written) ;
}

void OsInterpStackSizeExceeded (OsInterp interp) {
    (void) OsErrorTrigger (1, ose_stack) ;
}

void OsInterpClicksExceeded (OsInterp interp) {
    (void) OsErrorTrigger (1, ose_clicks) ;
}

void OsInterpSecurityFailed (OsInterp interp) {
    (void) OsErrorTrigger (1, ose_security) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (clone-interp SECURITY READL WRITEL CLICKSL)
 * Clone the current interpreter, limiting it with an amount of resources.
 * This primitive creates a new interpreter context. This new context gets
 * a copy of the source interpreter name space and macro space. This means
 * that atoms created in the clone will not be visible from the source.
 * Moreover, when "secure" is provided, the created interpreter will run scheme
 * code in a reduced environment, limited by:
 *     SECURITY: a security level, limiting access to scheme primitives,
 *     READL:    it will not be able to read more then READL from the file
 *               system,
 *     WRITEL:   it will not be able to write more then WRITEL byte to the
 *               local file system,
 *     CLICKSL:  it will not be able to consume more then CLICKSL clicks of the
 *               scheme virtual machine (roughly a clicks is equivalent to a
 *               virtual machine instruction execution).
 */

static OsObject OsSubrCloneInterp (OsInteger security
                                   , OsInteger os_read_limit
                                   , OsInteger os_write_limit
                                   , OsInteger os_clicks_limit) {
    OsInterp clone ;
    OsShouldBeInteger (security) ;
    OsShouldBeInteger (os_read_limit) ;
    OsShouldBeInteger (os_write_limit) ;
    OsShouldBeInteger (os_clicks_limit) ;
    /* For the time being, we use the default_stack_size, as we don't want to
     * bother scheme programmers with these nifty details */
    clone = OsInterpClone (OsCurrentInterp()
                           , default_stack_size
                           , security->ival
                           , (unsigned long) os_read_limit->ival
                           , (unsigned long) os_write_limit->ival
                           , (unsigned long) os_clicks_limit->ival) ;
    return ((OsObject) clone) ;
}

/* AUTODOC
 * (with-interp INTERP THUNK)
 * Execute THUNK in INTERP's context.
 * This primitive runs THUNK code block in the context of INTERP. THUNK is
 * invoked with the previous INTERP context as its only parameter. THUNK can
 * finish either:
 *     - beacuse of an error: in this case, the error is propagated back to
 *       the calling interpreter (which can not catch it, on purpose). If the
 *       calling interpreter has to do some cleanup after THUNK execution, it
 *       should use 'dynamic-wind with an approriate AFTER thunk.
 *     - with some result: in this case, the result is passed to the calling
 *       interpreter continuation, which is grabbed just before the
 *       'with-interp call.
 * Grabbing the calling interpreter continuation before running 'with-interp
 * allows the called interpreter to invoke continuations inside the calling
 * interpreter. This is particularly handy to simulate traps from the reduced
 * interpreter to the unreduced one.
 */

static OsObject OsContSubrWithInterp (OsContBlock cont
                                      , unsigned int argc
                                      , OsObject *argv) {
    OsInterp interp ;
    OsCodeBlock thunk ;
    OsInterp old_interp ;
    OsObject error = (OsObject) 0, value = (OsObject) 0 ;

    if ( argc != 2 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (2)
                               , OsIntegerMake (argc)) ;
    }
    interp = (OsInterp) argv[0] ;
    thunk  = (OsCodeBlock) argv[1] ;
    OsMustBeInterp (interp) ;
    OsMustBeClosedBlock (thunk) ;
    old_interp = OsInterpSwitch (interp) ;
    value = OsCodeBlockApplyWithArray (interp
                                       , thunk
                                       , 1, (OsObject *) &old_interp
                                       , &error) ;
    (void) OsInterpSwitch (old_interp) ;
    if ( error ) {
        os_uncatched_error (error) ;
    } else {
        OsCodeBlockApplyWithArray (old_interp
                                   , (OsCodeBlock) cont
                                   , 1, (OsObject *) &value
                                   , (OsObject *) 0) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (current-interp)
 * Returns the current interpreter context.
 */

static OsObject OsSubrCurrentInterp (void) {
    return ((OsObject) OsCurrentInterp()) ;
}

/* AUTODOC
 * (export-symbols LIST INTERP)
 * Export all symbols of LIST to INTERP.
 * This function recursively walks through LIST, exporting all encountered
 * symbols to INTERP. It returns a copy of the list, with all original symbols
 * changed to INTERP's symbols.
 */

static OsObject export_symbols (OsCons lst, OsInterp interp) {
    OsObject start = OsA_Nil, *plist = &start ;

    do {
        OsObject car = OS_CONS_CAR (lst) ;
        if ( OsIsAtom (car) ) {
            OsAtom exp = OsAtomMakeWithString (interp
                                               , ((OsAtom) car)->pname
                                               , OsA_Unbound) ;
            *plist = (OsObject) OsConsMake ((OsObject) exp, OsA_Nil) ;
            plist  = &(OS_CONS_CDR (*plist)) ;
        } else if ( OsIsCons (car) ) {
            OsObject  sublst = export_symbols ((OsCons) car, interp) ;
            *plist = (OsObject) OsConsMake (sublst, OsA_Nil) ;
            plist  = &(OS_CONS_CDR (*plist)) ;
        } else {
            *plist = (OsObject) OsConsMake (car, OsA_Nil) ;
            plist  = &(OS_CONS_CDR (*plist)) ;
        }
    } while ( (lst = (OsCons) OS_CONS_CDR (lst), OsIsCons (lst)) ) ;
    if ( lst != (OsCons) OsA_Nil ) {
        if ( OsIsAtom (lst) ) {
            *plist = (OsObject) OsAtomMakeWithString (interp
                                                      , ((OsAtom) lst)->pname
                                                      , OsA_Unbound) ;
        } else {
            *plist = (OsObject) lst ;
        }
    }
    return (start) ;
}

static OsObject OsSubrExportSymbols (OsCons lst, OsInterp interp) {
    OsShouldBeCons (lst) ;
    OsMustBeInterp (interp) ;
    return ((OsObject) export_symbols (lst, interp)) ;
}

/*
 * Initialize :
 */

void OsInterpInitialize (OsInterp interp, int *pargc, char **argv) {
    /* WARNING: interp not initialized yet ! */
    unsigned int argc = 1 ;

    OsInterpClass = OsClassDeclare ("Interp", sizeof (OsInterpRec)) ;
    
    while ( argc < *pargc ) {
        if ( ! strcmp ("-ss", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            default_stack_size = atoi (argv[argc]) ;
            OsArgvUsed (pargc, argc, argv) ;
#ifdef OS_SECURE
        } else if ( ! strcmp ("-clicks", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            default_clicks = (unsigned long) atol (argv[argc]) ;
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-read", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            default_read_bytes = (unsigned long) atol (argv[argc]) ;
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-write", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            default_written_bytes = (unsigned long) atol (argv[argc]) ;
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-chroot", argv[argc]) ) {
            OsArgvUsed (pargc, argc, argv) ;
            if ( chroot (argv[argc]) ) {
                fprintf (stderr, "oscheme: unable to chroot !") ;
                exit (1) ;
            }
            OsArgvUsed (pargc, argc, argv) ;
        } else if ( ! strcmp ("-security", argv[argc])) {
            OsArgvUsed (pargc, argc, argv) ;
            switch (*argv[argc]) {
              case 'a': default_security = OS_SECURITY_LEVEL_A ; break ;
              case 'b': default_security = OS_SECURITY_LEVEL_B ; break ;
              case 'c': default_security = OS_SECURITY_LEVEL_C ; break ;
              case 'd': default_security = OS_SECURITY_LEVEL_D ; break ;
              case 'e': default_security = OS_SECURITY_LEVEL_E ; break ;
              case 'f': default_security = OS_SECURITY_LEVEL_F ; break ;
              case '0':
              case '1':
              case '2':
              case '3':
              case '4':
              case '5':
              case '6':
              case '7':
              case '8':
              case '9':
                  default_security = atoi (argv[argc]) ; break ;
              default:
                  fprintf (stderr,"Invalid security level (%s)\n",argv[argc]);
                  exit (0) ;
            }
            OsArgvUsed (pargc, argc, argv) ;
#endif /* OS_SECURE */
        } else {
            argc++ ;
        }
    }
    return ;
}

void OsInterpSubrInitialize (OsInterp interp) {
    /* See comments in OsInterpMake() */
    interp->error_handler = OsA_False ;
    /* Set security errors */
    ose_clicks   = OsErrorMake (interp
                                , "error:clicks-exceeded"
                                , "Number of clicks exceeded"
                                , 0, OSE_CATCHABLE) ;
    ose_read     = OsErrorMake (interp
                                , "error:read-bytes-exceeded"
                                , "Number of read bytes exceeded"
                                , 0, OSE_CATCHABLE) ;
    ose_written  = OsErrorMake (interp
                                , "error:written-bytes-exceeded"
                                , "Number of written bytes exceeded."
                                , 0, OSE_CATCHABLE) ;
    ose_stack    = OsErrorMake (interp
                                , "error:stack-size-exceeded"
                                , "Stack size exceeded."
                                , 0, OSE_SEVERE) ;
    ose_security = OsErrorMake (interp
                                , "error:security-failed"
                                , "Security level too low."
                                , 0, OSE_CATCHABLE) ;
    OsSecSubrMake (interp, "clone-interp", OsSubrCloneInterp
                   , 4, OS_SECURITY_LEVEL_A) ;
    OsSecContSubrMake (interp, "with-interp", OsContSubrWithInterp
                       , 2, OS_SECURITY_LEVEL_A) ;
    OsSecSubrMake (interp, "export-symbols", OsSubrExportSymbols
                   , 2, OS_SECURITY_LEVEL_A) ;
    OsSubrMake (interp, "current-interp", OsSubrCurrentInterp, 0) ;
    return ;
}
