;;***************************************************************************
;; apropos.scm
;; apropos.scm,v 1.1.1.1 1995/02/27 15:16:08 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; On line help facilities. 
;; This module provides on-line help facilities. All the documentation is
;; extracted from the sources (C or scheme), and gathered by the autodir 
;; scheme script. 
;;***************************************************************************

;; Utilities:

;; Get all files in load-path having the given name:

(define (help:get-files filename)
  (let loop ((lp (load-path)) (files ()))
    (if (null? lp)
        files
        (let ((file (path-append (car lp) filename)))
          (loop (cdr lp) 
                (if (file-exists? file) (cons file files) files))))))

;; AUTODOC 
;; *help-with-mosaic*
;; Use the Mosaic browser to display help.
;; This variable control whether oscheme should use Mosaic to display on line
;; help. This requires:
;; - the process interface
;; - the mosaic interface
;; - runnin Mosaic 2.4 (at least) on the machine running oscheme
;; If all this holds, than you can set this toggle to true.

(define *help-with-mosaic* #f)

;; AUTODOC
;; *oscheme-help-url*
;; URL to contact to get oscheme help.
;; This symbol should be defined to the string describing the basename of
;; the URL to contact to get helpped by Mosaic.

(define *oscheme-help-url* "file://localhost//usr/local/lib/oscheme/html")

(define (help:with-mosaic sym)
  (let ((url (path-append *oscheme-help-url*
                          (string-append "help.html#" sym))))
    [(mosaic:make-pilot) goto: url]))

(define (help:module-with-mosaic module)
  (let ((url (path-append *oscheme-help-url*
                          (string-append "mindex.html#" module))))
    [(mosaic:make-pilot) goto: url]))

;; AUTODOC AUTOLOAD
;; (apropos STRING)
;; Find apropos documentation for every symbol matching regexp STRING.
;; This function browse all possible apropos files and tries to match the
;; given regular expression STRING against each encountered symbol name.
;; For all matching symbol, the apropos documentation is displayed to the
;; current output port.
;; The module in which the function is defined is display between brackets. 
;; This module name can be used through 'help-module, to get cross refs.

(define apropos ())

;; AUTODOC AUTOLOAD
;; (help STRING)
;; Display available help text for the given symbol.
;; This function displays to current output the help text for the given symbol.
;; If *help-with-mosaic* is set to #t, than this function will use Mosaic to
;; display the help, instead of the less convenienet current output.

(define help ())

(let ((help-re (regexp:make 
                (string-append "^\\*ENTRY\\*"
                               "[ \t]*([^ &\t]*)[ \t]*&" ;; symbol
                               "[ \t]*([^&]*)[ \t]*&" ;; call 
                               "[ \t]*([^ &\t]*)[ \t]*&" ;; module
                               "(.*)$"))))
  ;; Look given apropos file for sym
  (define (find-entry file re help-p)
    (with-input-from-file file
      (lambda ()
        (let loop ((line (read-line)) (indoc #f))
          (cond ((eof-object? line) #t)
                (indoc 
                 (if (regexp:match help-re line)
                     (loop line #f)
                     (begin (display line) (newline)
                            (loop (read-line) #t))))
                (else
                 (let ((match (regexp:match help-re line 1 2 3 4)))
                   (if (not match)
                       ;; skip unwanted doc
                       (loop (read-line) #f)
                       ;; test symbol, and display apropos and/or doc
                       (begin 
                         (if (regexp:match re (car match))
                             (begin (format (current-output-port) 
                                            "%0 [%1]\n\t%2\n%3\n" 
                                            (car match) 
                                            (car (cddr match))
                                            (cadr match)
                                            (cadr (cddr match)))
                                    (loop (read-line) help-p))
                             (loop (read-line) #f)))))))))))
  (define (doc string help-p)
    (let ((re (regexp:make string)))
      (for-each 
       (lambda (file) (find-entry file re help-p))
       (help:get-files "help.txt"))))
  (set! apropos (lambda (string) (doc string #f)))
  (set! help    
        (lambda (string) 
          (if *help-with-mosaic*
              (help:with-mosaic string)
              (doc (string-append "^" string "$") #t)))))

;; AUTODOC AUTOLOAD
;; (help-module MODULE)
;; Display documentation of module.
;; This function displays available documentation for MODULE, and all the 
;; functions defined by it.

(define help-module 
  (let ((split-re (regexp:make "[ \t]*([^& \t]*)[ \t]*&[ \t]*(.*)$"))
        (entry-re (regexp:make "\\*ENTRY\\*[ \t]*(.*)$")))
    (define (split-index-line line)
      (let split ((remains line) (ret ()))
        (let ((match (regexp:match split-re remains 1 2)))
          (if match 
              (split (cadr match) (cons (car match) ret))
              (reverse (cons remains ret))))))
    (define (do-documentation iport items)
      (format (current-output-port) "Module \"%0\" defines:\n" (car items))
      ;; Display defined symbols
      (for-each 
       (lambda (item) (format (current-output-port) "\t%0\n" item))
       (cdr items))
      ;; Display documentation
      (let docloop ((line (read-line iport)))
        (and (not (eof-object? line))
             (let ((description (regexp:match entry-re line 1)))
               (if description
                   (car description) ;; done
                   (begin (display line) (newline)
                          (docloop (read-line iport))))))))
    (define (display-module-documentation iport module) 
      (let loop ((line (read-line iport)))
        (or (eof-object? line)
            (let ((match (regexp:match entry-re line 1)))
              (and match (regexp:match module (car match))
                   (do-documentation iport (split-index-line line)))
              (loop (read-line iport))))))
    (lambda (module)
      (if *help-with-mosaic*
          (help:module-with-mosaic module)
          (let ((module-re (regexp:make module)))
            (for-each 
             (lambda (file) 
               (display-module-documentation (open-input-file file) module-re))
             (help:get-files "index.txt")))))) )

(provide "apropos")
