;; date.scm
;; date.scm,v 1.1.1.1 1995/02/27 15:16:09 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; Various date handling utilities.
;; This file is merely a scheme version of the file calendar.el by M. Reingold.
;; Date are objects having the following slots:
;; year, month, day, hour, minute, second
;; And the following methods:
;; inDays, inSeconds

;; AUTODOC AUTOLOAD
;; (date:leap-year-p YEAR)
;; Is YEAR a leap year in the gregorian calendar.

(define (date:leap-year-p year)
  (and (= (remainder year 4) 0)
       (or (not (= 0 (remainder year 100)))
           (= 0 (remainder year 400)))))

;; AUTODOC AUTOLOAD
;; (date:last-day-of-month MONTH YEAR)
;; The last day in MONTH during YEAR.

(define date:last-day-of-month 
  (let ((d-in-m (vector 31 28 31 30 31 30 31 31 30 31 30 31)))
    (lambda (month year)
      (if (and (= month 2) (date:leap-year-p year))
          29
          (vector-ref d-in-m (- month 1))))) )

;; AUTODOC AUTOLOAD
;; (date:day-number DATE)
;; Returns the day number within the year of the date DATE.
 
(define (date:day-number date)
  (let* ((month [date month])
         (day   [date day])
         (year  [date year])
         (d-o-y (+ day (* 31 (- month 1)))))
    (and (> month 2)
         (begin
           (set! d-o-y (- d-o-y (quotient (+ 23 (* 4 month)) 10)))
           (and (date:leap-year-p year) (set! d-o-y (+ 1 d-o-f)))))
    d-o-y) )

;; AUTODOC AUTOLOAD
;; (date:make-date YEAR MONTH DAY [HOUR MINUTE SECOND])
;; Create a date object holding the given date.

(define date:make-date 
  (let ((date [obj:root clone]))
    [date defineSlot: "year"   init: 1]
    [date defineSlot: "month"  init: 12]
    [date defineSlot: "day"    init: 31]
    [date defineSlot: "hour"   init: 0]
    [date defineSlot: "minute" init: 0]
    [date defineSlot: "second" init: 0]
    ;; Display a date:
    (define-message date (display)
      (format (current-output-port)
              "%0/%1/%2 %3:%4:%5"
              [self day] [self month] [self year]
              [self hour] [self minute] [self second]))
    ;; Convert date in days since 21/31/1
    (define-message date (inDays)
      (let ((prior-years (- [self year] 1)))
        (+ (date:day-number self)
           (* 365 prior-years)
           (quotient prior-years 4)
           (- (quotient prior-years 100))
           (quotient prior-years 400))))
    ;; Convert date to a # of seconds since 1/12/70:
    (define-message date (inSeconds)
      (let* ((days-in-1970 [(date:make-date 1970 1 1) inDays])
             (days (- [self inDays] days-in-1970)))
          (+ (* days 24 60 60)
             (* [self hour] 60 60)
             (* [self minute] 60)
             [self second])))
    ;; Save the date to some file (to be read through date:read-date)
    (define-message date (asVector)
      (vector [self year] [self month] [self day]
              [self hour] [self minute] [self second]))
    ;; Format given date:
    (let ((days-name (vector "Sunday" "Monday" "Tuesday" "Wednesday" 
                             "Thursday" "Friday" "Saturday"))
          (months-name (vector "January" "February" "March" "April" "May" 
                               "June" "July" "August" "September" "October" 
                               "November" "December")))
      (define-message date (format)
        (format (current-output-port)
                "%0, %1 %2 %3 (%4:%5:%6)"
                (vector-ref days-name (modulo [self inDays] 7))
                [self day]
                (vector-ref months-name (- [self month] 1))
                [self year]
                [self hour] [self minute] [self second])))
    ;; Compare with another date:
    (define-message date (compare: anotherDate)
      (if (and (>= [self year] 1970) (>= [anotherDate year] 1970))
          (- [self inSeconds] [anotherDate inSeconds])
          (- [self inDays] [anotherDate inDays])))
    (lambda (year month day . rest)
      (let ((new [date clone]))
        [new year: year]
        [new month: month]
        [new day: day]
        (and (pair? rest) [new hour: (car rest)]
             (pair? (cdr rest)) [new minute: (cadr rest)]
             (pair? (cddr rest)) [new second: (car (cddr rest))])
        new))) )
        
;; AUTODOC AUTOLOAD
;; (date:days-to-date DAYS)
;; Return a date object for DAYS after 12/31/1.

(define (date:days-to-date days)
  (let* ((d0   (- days 1))
         (n400 (quotient d0 146097))
         (d1   (modulo d0 146097))
         (n100 (quotient d1 36524))
         (d2   (modulo d1 36524))
         (n4   (quotient d2 1461))
         (d3   (modulo d2 1461))
         (n1   (quotient d3 365))
         (day  (+ (modulo d3 365) 1))
         (year (+ (* 400 n400) (* 100 n100) (* n4 4) n1)))
    (if (or (= n100 4) (= n1 4))
        (date:make-date year 12 31)
        (let ((year (+ 1 year)))
          (let loop ((month 1))
            (let ((mdays (date:last-day-of-month month year)))
              (if (and (< mdays day) (set! day (- day mdays)))
                  (loop (+ month 1))
                  (date:make-date year month day))))))) )

;; AUTODOC AUTOLOAD
;; (date:seconds-to-date SECONDS)
;; Return a date object for SECONDS after 12/1/70, 00:00:00

(define date:seconds-to-date 
  (let ((days-in-1970 [(date:make-date 1970 1 1) inDays]))
    (lambda (secs)
      (let ((days (quotient secs (* 24 60 60)))
            (time (remainder secs (* 24 60 60))))
        (let ((date (date:days-to-date (+ days days-in-1970))))
          (let* ((hours (quotient time (* 60 60)))
                 (ms    (remainder time (* 60 60)))
                 (minutes (quotient ms 60))
                 (seconds (remainder ms 60)))
            [date hour: hours]
            [date minute: minutes]
            [date second: seconds]
            date))))) )

;; AUTODOC AUTOLOAD
;; (date:vector-to-date INPUT-STREAM)
;; Convert a vector to a date.
;; This function should be used to save a date into some output stream. Use the
;; the asVector message of date, save the vector. Then, read back the vector
;; and convert it into a date by invoking this function.

(define (date:vector-to-date vect)
  (date:make-date (vector-ref vect 0)
                  (vector-ref vect 1)
                  (vector-ref vect 2)
                  (vector-ref vect 3)
                  (vector-ref vect 4)
                  (vector-ref vect 5)))

(provide "date")  

