;;***************************************************************************
;; debugger.scm
;; debugger.scm,v 1.2 1995/03/17 14:24:15 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; The debugger module provides a starting point for writing a debugger. It
;; can still help in its actual state to locate errors. Support is provided
;; by the kernel to write a much more complete debugger (see for example
;; the full documentation on breakpoints).
;; To run the debugger:
;; - require this module,
;; - require the "error" module,
;; - set the *debug-on-error* flag to #t
;; - compile your code with the compiler:debug flag set to #t
;;***************************************************************************  

(require "error")

(set! *debug-on-error* #t)

;; Display the call stack provided by the kernel

(define (display-call-stack vm-ctxt frameno)
  (let ((s (vector-length vm-ctxt)))
    (let loop ((i 0))
      (and (< i s)
           (let ((frame (vector-ref vm-ctxt i)))
             ;; display current expr 
             (format (current-output-port) 
                     "%0 %1 %2\n" 
                     (if (= frameno i) "*" " ") i (vector-ref frame 0))
             (loop (+ i 1)))))) )

;; Safe execute
;; Execute given expression, disarming the debugger if needed:

(define (debuger-undebug-eval expr)
  (let ((compiler:debug? compiler:debug)
        (debug-on-error? *debug-on-error*)
        (errors?         'debug-no-error))
    (let ((value (error:case 
                  (let ((block (sys:compile (sys:expand expr)))) 
                    (codeblock:execute block))
                  (else (lambda errors (set! errors? errors))))))
      (set! compiler:debug   compiler:debug?)
      (set! *debug-on-error* debug-on-error?)
      (if (eq? errors? 'debug-no-error) value (error:format errors?)))))

;; Debuger's main loop (terse ;-)

(define (debuger:repl vm-ctxt errors)
  (let ((exit? #f) 				;; Should repl exit
        (exit-value (unbound))			;; returned value
        (fidx 0)				;; current frame #
        (fmax (vector-length vm-ctxt)))		;; max # of frames
    ;; Display a warning:
    (define (debuger:warning . rest)
      (apply format (current-output-port) rest)
      (output-port-flush (current-output-port)))
    ;; Display help
    (define (debuger:help)
      (display "up     : goes up one frame\n")
      (display "down   : goes down one frame\n")
      (display "frame  : go to given FRAME\n")
      (display "locals : show local variables for current frame\n")
      (display "where  : show full call stack\n")
      (display "quit   : exit debugger, back to toplevel\n")
      (display "error  : display error that trigerred the debugger\n")
      (display "return : return from top frame with VALUE\n")
      (display "help   : this help\n")
      (display "All commands can be invoked by their first letter\n")
      (output-port-flush (current-output-port)))
    ;; Read a command from stdin, for the debugger
    (define (debuger:read)
      (format (current-output-port) "(%0)? " fidx)
      (output-port-flush (current-output-port))
      (read))
    ;; Up one frame
    (define (debuger:up)
      (and (< (+ fidx 1) fmax) (set! fidx (+ fidx 1))))
    ;; Down one frame
    (define (debuger:down)
      (and (>= (- fidx 1) 0) (set! fidx (- fidx 1))))
    ;; Show local vars of current frame
    (define (debuger:locals)
      (let ((env (vector-ref (vector-ref vm-ctxt fidx) 1)))
        (display env) (newline)))
    ;; Where are we: full stack dump
    (define (debuger:where)
      (display-call-stack vm-ctxt fidx))
    ;; Go to given frame
    (define (debuger:frame)
      (let ((nfidx (read)))
        (if (and (>= nfidx 0) (< nfidx fmax))
            (set! fidx nfidx)
            (debuger:warning "invalid frame number %0\n" nfidx))))
    (define (debuger:print)
      (define (make-expr expr env)
        (let loop ((renv env) (letenv ()))
          (if (null? renv)
              (cons* 'let letenv (list expr))
              (let loopi ((bp (car renv)) (lp ()))
                (if (null? bp)
                    (loop (cdr renv) (append lp letenv))
                    (loopi (cddr bp) 
                           (cons (list (car bp)
                                       (list 'quote (cadr bp))) lp)))))))
      (let ((env (vector-ref (vector-ref vm-ctxt fidx) 1)))
        (if (pair? env)
            (let ((expr (make-expr (read) env)))
              (debuger:warning "returned value: %0\n"
                               (debuger-undebug-eval expr)))
            (begin (read) (debuger:warning "No env for current frame\n")))) )
    ;; main loop
    (let loop ((cmd (debuger:read)))
      (case cmd
        ((up u) (debuger:up))
        ((down d) (debuger:down))
        ((frame f) (debuger:frame))
        ((locals l) (debuger:locals))
        ((where w) (debuger:where))
        ((quit q) (set! exit? #t))
        ((error e) (error:format errors))
        ((return r) (set! exit-value (read)) (set! exit? #t))
        ((help h ?) (debuger:help))
        ((print p) (debuger:print))
        (else (debuger:warning "Invalid command %0\n" cmd)))
      (or (and exit? exit-value)
          (loop (debuger:read))))) )
    
(define (debuger:enter errors)
  (let ((ctxt (memq error:vm-context errors)))
    (if ctxt
        (let ((vm-ctxt (car (cadr ctxt))))
          (display-call-stack vm-ctxt 0)
          (debuger:repl vm-ctxt errors))
        (unbound))) )

(provide "debugger")
