;;***************************************************************************
;; error.scm
;; error.scm,v 1.2 1995/03/12 02:58:48 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;;***************************************************************************

(require "autoload")

;; AUTODOC
;; *debug-on-error*
;; Set automatic entry into debugger when an error occurs.
;; When set to #t, this variable will trigger entry into the debugger when any
;; error occurs (even if it is an uncatchable error).

(define *debug-on-error* #f)

;; Simple print of current error, skiping the vm state, which might
;; garble thee screen.

(define (error:format errors)
  (let loop ((e errors))
    (and (pair? e)
         (begin 
           (and (not (eq? (car e) error:vm-context))
                (begin
                  (apply format 
                         (current-output-port)
                         (error->format (car e))
                         (cadr e))
                  (newline (current-output-port))))
           (loop (cddr e))))) )

;; Raw error handler (called by the kernel)

(define (error:default-handler . errors)
  (or 
   ;; Test for an autoloadable unbound atom:
   (and (eq? (car errors) error:unbound)
        (let ((symbol (autoload:unbound-corrector (car (cadr errors)))))
          (and (bound? symbol) symbol)))
   ;; Should we enter the debugger ?
   (and *debug-on-error* 
        (let ((retval (debuger:enter errors)))
          (and (bound? retval) retval)))
   ;; We can't do anything to help !
   (unbound)))

;; AUTODOC AUTOLOAD
;; (error:case EXPR CLAUSES)
;; Evaluate EXPR, cathing errors.
;; This macro expands itself to a form that will evaluate EXPR. If some error
;; occurs during this evaluation, it is caught and match against the CLAUSES.
;; Each clause in CLAUSES should have one of the following forms:
;;     ((ERROR1 ERROR2 ...ERRORN) HANDLER)
;;     (else HANDLER)
;; If an error is matched against the error that occured, than the 
;; corresponding HANDLER is invoked with the error (and its associated 
;; context). 
;; Clauses of the second form match any errors.

(define (error:expand-case-clauses error-variable error-context clauses)
  (define (expand-one clause)
    (if (or (eq? (car clause) #t) (eq? (car clause) 'else))
        `(#t (apply ,(cadr clause) ,error-context))
        `((memq ,error-variable (list ,@(car clause)))
          (apply ,(cadr clause) ,error-context))))
  (define (aux clauses expansion)
    (if (null? clauses)
        (reverse expansion)
        (aux (cdr clauses)
             (cons (expand-one (car clauses)) expansion))))
  `(cond ,@(aux clauses ())))

(define-macro (error:case call)
  (let ((error-name (gensym)))
    `(let ((escape ())
           (old-handler #f))
       (dynamic-wind
        (lambda () 
          (or old-handler (set! old-handler (error-handler))))
        (lambda (f)
          (set! escape f)
          (set-error-handler!
           (lambda ,error-name
             ;; Set back old error handler, if some error occurs now:
             (set-error-handler! old-handler)
             (escape ,(error:expand-case-clauses 
                       `(car ,error-name)
                       error-name 
                       (cddr call)))))
          ,(cadr call))
        (lambda () (set-error-handler! old-handler))))) )
 
(or (error-handler) (set-error-handler! error:default-handler))

(provide "error")
