;;***************************************************************************
;; expand.scm
;; expand.scm,v 1.2 1995/03/14 11:52:18 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; The expand modules defines all R4RS syntactic construction, by using the
;; macro facilities provided by oscheme.
;;***************************************************************************

;; AUTODOC
;; (letrec BINDING BODY)
;; Evaluate the BODY after having set BINDING recursively.
;; The variables are bound to fresh locations holding undefined values, the
;; inits are evaluated in the resulting environment (in some unspecified
;; order), each variable is assigned to the result of the corresponding init,
;; the body is evaluated in the resulting environment, and the value of the
;; last expression in body is returned.  Each binding of a variable has the
;; entire letrec expression as its region region, making it possible to define
;; mutually recursive procedures.
;; Example:
;; (letrec ((even?
;;           (lambda (n)
;;             (if (zero? n)
;;                 #t
;;                 (odd? (- n 1)))))
;;          (odd?
;;           (lambda (n)
;;             (if (zero? n)
;;                 #f
;;                 (even? (- n 1))))))
;;   (even? 88))   
;; = #t

(define (expand-letrec bindings body)
  `((lambda ,(map (lambda (x) (if (pair? x) (car x) x)) bindings)
      ,@(map 
         (lambda (x) (if (pair? x) `(set! ,(car x) ,(cadr x)) `(set! ,x ())))
         bindings) ,@body)
    ,@(map (lambda (x) '(unbound)) bindings)))

(define-macro (letrec call)
  (expand-letrec (cadr call) (cddr call)) )

;; let:
;; The named let version requires one more closure. This is unavoidable as
;; the named-function might be used as a return value (sighh)
;; So, better use not-named lambda when ever you can
;; AUTODOC
;; (let [NAME] BINDINGS BODY)
;; Evaluate BODY after having set the BINDINGS.
;; Bindings should be a list of the following forms:
;; (VARIABLE INIT)
;; where each INIT is an expression.  It is an error for a variable to appear
;; more than once in the list of variables being bound.
;; The inits are evaluated in the current environment (in some unspecified 
;; order), the variables are bound to fresh locations holding the results, the
;; body is evaluated in the extended environment, and the value of the last 
;; expression of body is returned.
;; Named let has the same syntax and semantics as ordinary let except that
;; variable is bound within body to a procedure whose formal arguments are the
;; bound variables and whose body is body.  Thus the execution of body may be
;; repeated by invoking the procedure named by variable.
;; Example:
;; (let ((x 2) (y 3))
;;   (* x y))                      
;; = 6
;; (let ((x 2) (y 3))
;;   (let ((x 7)
;;         (z (+ x y)))
;;     (* z x)))                   
;; = 35

(define (expand-unamed-let bindings body)
  `((lambda ,(map (lambda (x) (if (pair? x) (car x) x)) bindings) ,@body)
    ,@(map (lambda (x) (if (pair? x) (cadr x) ())) bindings)) )

(define (expand-named-let name bindings body)
  `(letrec ((,name (lambda 
                       ,(map (lambda (x) (if (pair? x) (car x) x)) bindings) 
                     ,@body)))
     (,name ,@(map (lambda (x) (if (pair? x) (cadr x) ())) bindings))))

(define-macro (let call)
  (if (symbol? (cadr call))
      (expand-named-let (cadr call) (car (cddr call)) (cdr (cddr call)))
      (expand-unamed-let (cadr call) (cddr call))) )

;; cond:
;; lacks a check to make sure there is only one else clause...
;; AUTODOC
;; (cond CLAUSES)
;; Evaluates the first CLAUSE that is #t.
;; Each CLAUSES should be of the form:
;; (TEST EXPRESSION*)
;; The last clause may be an "else" clause, which has the form:
;; (ELSE EXPRESSION*)
;; A cond expression is evaluated by evaluating the TEST EXPRESSIONs of
;; successive CLAUSES in order until one of them evaluates to a true value true
;; (see section booleansection).  When a TEST evaluates to a true value, then
;; the remaining EXPRESSIONs in its clause are evaluated in order, and the
;; result of the last EXPRESSION in the clause is returned as the result of the
;; entire cond EXPRESSION.  If the selected clause contains only the TEST and 
;; no EXPRESSIONs, then the value of the TEST is returned as the result.  If 
;; all TESTs evaluate to false values, and there is no else clause, then the 
;; result of the conditional EXPRESSION is unspecified; if there is an else 
;; clause, then its EXPRESSIONs are evaluated, and the value of the last one is
;; returned.
;; Examples:
;; (cond ((> 3 2) 'greater)
;;       ((< 3 2) 'less))         greater
;; (cond ((> 3 3) 'greater)
;;       ((< 3 3) 'less)
;;       (else 'equal))            equal


(define (expand-cond-clause clause)
  `(if ,(if (eq? (car clause) 'else) #t (car clause))
       (begin ,@(cdr clause))))

(define (expand-cond-clauses clauses)
  (if (null? clauses)
      #f
      (append (expand-cond-clause (car clauses))
              (list (expand-cond-clauses (cdr clauses))))) )

(define (expand-cond-clauses clauses)
  (if (null? clauses)
      #f
      (append (expand-cond-clause (car clauses))
              (list (expand-cond-clauses (cdr clauses))))))

(define-macro (cond call) (expand-cond-clauses (cdr call)))

;; AUTODOC
;; (case KEY CLAUSE)
;; Evaluates the CLAUSE matching KEY.
;; Key may be any expression.  Each clause should have the form:
;; ((DATUM*) EXPRESION*)
;; where each DATUM is an external representation of some object.  All the
;; DATUMs must be distinct.  The last clause may be an ``else clause,'' which
;; has the form:
;; (else EXPRESSION*)
;; A case expression is evaluated as follows.  Key is evaluated and its result
;; is compared against each DATUM.  If the result of evaluating KEY is
;; equivalent (in the sense of eqv?; see section eqv?) to a DATUM, then the
;; expressions in the corresponding clause are evaluated from left to right and
;; the result of the last expression in the clause is returned as the result of
;; the case expression.  If the result of evaluating KEY is different from
;; every DATUM, then if there is an else clause its expressions are evaluated
;; and the result of the last is the result of the case expression; otherwise
;; the result of the case expression is unspecified.  
;; Examples:
;; (case (* 2 3)
;;    ((2 3 5 7) 'prime) ((1 4 6 8 9) 'composite)) 
;; = composite 
;; (case (car '(c d))
;;   ((a) 'a) ((b) 'b)) (case (car '(c d)) ((a e i o u) 'vowel) 
;;   ((w y) 'semivowel) (else 'consonant)) 
;; = consonant 


(define (expand-case-clauses syme clauses)
  (if (null? clauses)
      '(unbound)
      (let ((datum (caar clauses)))
        (if (eq? datum 'else)
            `(if #t (begin ,@(cdar clauses)) #f)
            `(if (memq ,syme (quote ,datum))
                 (begin ,@(cdar clauses))
                 ,(expand-case-clauses syme (cdr clauses)))))) )

(define (expand-case expr clauses)
  (let ((syme (gensym "case")))
    `((lambda (,syme) ,(expand-case-clauses syme clauses)) ,expr)))

(define-macro (case call) (expand-case (cadr call) (cddr call)))

;; let*
;; AUTODOC
;; (let* BINDINGS BODY)
;; Evaluates BODY after having set the BINDINGS sequentially.
;; Let* is similar to let, but the bindings are performed sequentially from
;; left to right, and the region region of a binding indicated by (variable
;; init) is that part of the let* expression to the right of the binding.  Thus
;; the second binding is done in an environment in which the first binding is
;; visible, and so on.
;; Example:
;; (let ((x 2) (y 3))
;;   (let* ((x 7)
;;         (z (+ x y)))
;;    (* z x)))             
;; = 70

(define (expand-let* bindings body)
  (cond ((null? bindings) `((lambda () ,@body)))
        ((null? (cdr bindings))
         (let ((binding (car bindings)))
           (if (pair? binding)
               `((lambda (,(car binding)) ,@body) ,(cadr binding))
               `((lambda (,binding) ,@body) ()))))
        (else
         (let ((binding (car bindings)))
           (if (pair? binding)
               `((lambda (,(car binding)) ,(expand-let* (cdr bindings) body))
                 ,(cadr binding))
               `((lambda (,binding) ,(expand-let* (cdr bindings) body))
                 ())  ))) ))

(define-macro (let* call)
  (expand-let* (cadr call) (cddr call)))

;; AUTODOC
;; (do ((VARIABLE INIT STEP)*) (TEST EXPRESSION) COMMAND*)
;; Iterating construct.
;; Do is an iteration construct.  It specifies a set of VARIABLEs to be bound,
;; how they are to be INITialized at the start, and how they are to be updated
;; on each iteration.  When a termination condition is met, the loop exits with
;; a specified result value.
;; Do expressions are evaluated as follows: The INIT expressions are evaluated
;; (in some unspecified order), the VARIABLEs are bound to fresh locations, the
;; results of the INIT expressions are stored in the bindings of the VARIABLEs,
;; and then the iteration phase begins.
;; Each iteration begins by evaluating test; if the result is false (see
;; section booleansection), then the command expressions are evaluated in order
;; for effect, the step expressions are evaluated in some unspecified order,
;; the VARIABLEs are bound to fresh locations, the results of the steps are
;; stored in the bindings of the VARIABLEs, and the next iteration begins.
;; If test evaluates to a true value, then the expressions are evaluated from
;; left to right and the value of the last expression is returned as the value
;; of the do expression.  If no expressions are present, then the value of the
;; do expression is unspecified.
;; The region region of the binding of a VARIABLE consists of the entire do
;; expression except for the INITs.  It is an error for a VARIABLE to appear
;; more than once in the list of do VARIABLEs.
;; A step may be omitted, in which case the effect is the same as if (VARIABLE
;; INIT VARIABLE) had been written instead of (VARIABLE INIT).
;; Example:
;; (do ((vec (make-vector 5))
;;      (i 0 (+ i 1)))
;;     ((= i 5) vec)
;;   (vector-set! vec i i))       
;; = (0 1 2 3 4)
;; (let ((x '(1 3 5 7 9)))
;;   (do ((x x (cdr x))
;;        (sum 0 (+ sum (car x))))
;;       ((null? x) sum))) 
;; = 25

(define (expand-do inits exprs cmds)
  (let ((sym (gensym)))
    `(let ,sym ,(map (lambda (i) (list (car i) (cadr i))) inits)
          (if ,(car exprs) 
              (begin ,@(cdr exprs))
              (begin ,@cmds
                     (,sym 
                      ,@(map
                         (lambda (i) 
                           (if (pair? (cddr i)) (car (cddr i)) (car i)))
                         inits)))))) )

(define-macro (do call) 
  (expand-do (cadr call) (car (cddr call)) (cdr (cddr call))))
  
;; lambda
;; Expands internal definitions

(define (massage-internals exprs idefs ibody)
  (if (null? exprs)
      (and (pair? idefs) (cons idefs ibody))
      (let ((e (car exprs)))
        (if (and (pair? e) (eq? (car e) 'define))
            (let ((f (cadr e)) (v (cddr e)))
              (massage-internals 
               (cdr exprs)
               (append idefs
                       (if (pair? f)
                           `((,(car f) (lambda ,(cdr f) ,@v)))
                           `((,f ,@v))))
               ibody))
            (massage-internals (cdr exprs) idefs (append ibody (list e)))))) )

(define-macro (lambda call)
  (let ((formals (cadr call)) (body (cddr call)))
    (let ((internals (massage-internals body () ())))
      (if internals 
          `(lambda ,formals (letrec ,(car internals) ,@(cdr internals)))
          call))) )

;; define
;; Expands implicit lambdas

(define-macro (define call)
  (let ((f (cadr call)) (v (cddr call)))
    (if (pair? f)
        `(define ,(car f) (lambda ,(cdr f) ,@v))
        call)) )

;; AUTODOC 
;; (define-message OBJECT HEADER BODY)
;; Refine OBJECT with a new message implementation.
;; OBJECT should be an external object, HEADER should have the form:
;; (SEL: VARIABLE ... [SEL])
;; This macro refines the implementation of the message whose name is the 
;; concatenatio nof all SEL parts in HEADER, with the new implementation
;; described by BODY. BODY is evaluated with all VARIABLE bound to their
;; corresponding parameter, plus the leading receiver (i.e self).
;; Example:
;; (define object [obj:root clone])
;; (define-message object (print: this to: stream)
;;     (format (current-output-port) "%0 print %1 to %2\n" self this stream))
;; [object print: 1 to: "foo"]

(define (expand-define-message object header body)
  (define (split h sels args)
    (cond ((null? h) (cons sels (reverse args)))
          ((null? (cdr h)) 
           (cons (string-append sels (symbol->string (car h))) (reverse args)))
          (else (split (cddr h)
                       (string-append sels (symbol->string (car h)))
                       (cons (cadr h) args)))))
  (let ((splitten (split header "" ())))
    `[,object defineMessage: (obj:define-selector ,(car splitten))
              to: (lambda ,(cons 'self (cdr splitten)) ,@body)]))

(define-macro (define-message call)
  (expand-define-message (cadr call) (car (cddr call)) (cdr (cddr call))))
 
;; AUTODOC AUTOLOAD
;; (when CONDITION BODY)
;; If CONDITION evaluates to true, then execute BODY.

(define-macro (when call)
  `(if ,(cadr call) (begin ,@(cddr call)) #f))

;; AUTODOC AUTOLOAD
;; (unless CONDITION BODY)
;; If CONDITION evaluates to false, then execute BODY.

(define-macro (unless call)
  `(if ,(cadr call) #t (begin ,@(cddr call))) )

(provide "expand")
