;;***************************************************************************
;; ftp package
;; ftp.scm,v 1.1.1.1 1995/02/27 15:16:06 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; A simple interface to ftp. 
;; This module provides a simple scheme interface to ftp. It is implemented
;; by objects (in fact, its starting point was to test the object model).
;***************************************************************************

(require "process")

;; Customizable part

(define ftp:command-line (list "ftp" "-i" "-g" "-v" "-n"))

;; Hardly customizable part:

(define ftp:skip-re 
  (regexp:make 
   (string-append 
    "^200 (PORT|Port) |^331 |^150 |^350 |^[0-9]+ bytes |"
    "^Connected |^$|^Remote system|^Using|^ |Password:|"
    "^local:|^Trying|^125 |^550-|^221 .*oodbye|^502 ")) )

(define ftp:fatal-re
  (regexp:make
   (string-append
    "^ftp: |^Not connected|^530 |^4[25]1 |rcmd: |"
    "^No control connection|unknown host|^lost connection")) )

(define ftp:size-re
  (regexp:make "^150 .* connection for .* (([0-9]+) bytes)"))

(define ftp:good-re 
  (regexp:make 
   "^220 |^230 |^226 |^25. |^221 |^200 |^[Hh]ash mark"))

(define ftp:multi-re
  (regexp:make
   "^220-|^230-|^226|^25.-|^221-|^200-|^530-|^4[25]1-"))

;; AUTODOC
;; (make-ftp)
;; This object provides a scheme interface to ftp protocol.
;; Launches an ftp client process. It has the following methods:
;; send-command:        sends given string command to underlying process,
;; wait-for-completion: wait for last command to terminate,
;; login:user:passwd:   login given host,
;; ls:                  list current directory,
;; cd:                  go to given directory,
;; And the following slots:
;; ftp-error    provides the last (fatal ?) encountered error
;; debug	trace dialog with underlying process if true
;; host 	current connected host
;; user		current account
;; passwd	current passwd
;; For sample usages, see the 'ftpget script.

(define make-ftp
  (let ((ftp [obj:root clone]))
    [ftp defineSlot: "busy" init: #f]
    [ftp defineSlot: "ftp-error" init: #f]
    [ftp defineSlot: "result" init: ""]
    [ftp defineSlot: "debug" init: #f]
    [ftp defineSlot: "host" init: ""]
    [ftp defineSlot: "user" init: ""]
    [ftp defineSlot: "passwd" init: ""]
    [ftp defineSlot: "state" init: ()]
    [ftp defineSlot: "multi-skip" init: #f]
    [ftp defineSlot: "process" init: #f]
    ;; This one tests the super construction:
    ;; It forks a new ftp process on clone, which is perhaps not the best
    ;; thing to do...
    (define-message ftp (clone)
      (let ((clone [[self super] clone])
            (proc  (apply make-process ftp:command-line)))
        [clone process: proc]
        clone))
    ;; Send a command to underlying ftp porcess
    ;; Caller should use wait-for-completion to wait for ftp process to 
    ;; perform its job.
    (define-message ftp (send-command: cmd)
      (if [self ftp-error]
          (error "ftp error \"%0\"" [self ftp-error])
          (begin [self busy: #t]
                 [self result: ""]
                 (display (string-append cmd "\n") [[self process] input])
                 (output-port-flush [[self process] input])
                 (and [self debug] 
                      (begin (display "<- ") (display cmd) (newline)))
                 #t)) )
    ;; wait-for-completion
    ;; Wait for last send-command to acheive its work.
    (define-message ftp (wait-for-completion)
      ;; Check for some pending data on the error stream of ftp client
      (let* ((eport [[self process] error])
             (ready (vector-ref (select 0 eport) 0)))
        (and (not (null? ready))
             (error "ftp client error \"%0\"\n" 
                    [self ftp-error: (read-line eport)])))
      ;; Okay, we can continue:
      (let loop ((line (read-line [[self process] output])))
        (and [self debug] (begin (display "-> ") (display line) (newline)))
        (cond ((regexp:match ftp:skip-re line) 
               (and [self debug] (display "SKIP-RE\n"))
               [self multi-skip: #f]
               (loop (read-line [[self process] output])))
              ((regexp:match ftp:good-re line)
               (and [self debug] (display "GOOD-RE\n"))
               [self multi-skip: #f]
               [self busy: #f]
               [self result: line]
               #t)
              ((regexp:match ftp:fatal-re line) 
               (and [self debug] (display "FATAL-RE\n"))
               [self busy: #f]
               [self multi-skip: #f]
               [self state: 'killed]
               [self ftp-error: line] 
               #f)
              ((regexp:match ftp:multi-re line) 
               (and [self debug] (display "MULTI-RE\n"))
               [self multi-skip: #t]
               (loop (read-line [[self process] output])))
              ([self multi-skip] 
               (and [self debug] (display "MULTI-RE cont'ed\n"))
               (loop (read-line [[self process] output])))
              (else (and [self debug] (display "RESULT\n"))
                    [self multi-skip: #f]
                    [self result: line]
                    #t))))
    ;; login:user:passwd
    (define-message ftp (login: host user: user passwd: pswd)
      (let ((ret [self send-command: (string-append "open " host)]))
        (and ret [self wait-for-completion] [self state: 'connected]))
      (let ((ret [self send-command: (string-append "user " user
                                                    " passwd \"" pswd "\"")]))
        (if ret
            (and [self wait-for-completion]
                 (begin [self host: host]
                        [self user: user]
                        [self passwd: pswd]
                        [self state: 'logged]))
            [self state: 'not-connected])))
    ;; ls
    ;; drwxr-xr-x   4 baird    users        3072 Jun  8 21:30 Texts
    (let ((splitre (regexp:make "\([^ \t]*\)[ \t]+\(.*\)"))
          (entry [obj:root clone]))
      ;; Entry is equiv of stat for local filesystem
      [entry defineSlot: "type"]
      [entry defineSlot: "filename"]
      [entry defineSlot: "size"]
      ;; goodies:
      (define-message entry (display)
        (format (current-output-port) 
                "#[ftp-entry %0 (%1)]" [self filename] [self size]))
      (define-message ftp (parse-ls: line) 
	;; Split a line into its components, returns a vector
	(define (split-line line)
	  (let ((vline (vector)))
	    (let loop ((rest line) (m (regexp:match splitre line 1 2)))
	      (if m
		  (begin (vector-append! vline (car m))
			 (loop (cadr m) (regexp:match splitre (cadr m) 1 2)))
		  (vector-append! vline rest)))))
        ;; Make an object from the vector:
        (let ((fields (split-line line))
              (ent [entry clone]))
          (and (> (vector-length fields) 8)
               (begin
                 ;; find  back the type
                 (case (string-ref (vector-ref fields 0) 0)
                   ((#\d) [ent type: 'directory])
                   ((#\l) [ent type: 'link])
                   ((#\-) [ent type: 'file])
                   (else  [ent type: 'unknown]))
                 ;; the size
                 [ent size: (string->integer (vector-ref fields 4))]
                 ;; name
                 [ent filename: (vector-ref fields 8)]
                 ent)))))
    (define-message ftp (ls)
      (and [self send-command: "dir"]
           [self wait-for-completion] 	;; skip the 'total ...' first line
           (let loop ((files ()))
             (and [self wait-for-completion]
                  (if [self busy]
                      (loop (cons [self parse-ls: [self result]] files))
                      (reverse files))))))
    ;; Get a file
    (define-message ftp (get: filename)
      (and [self send-command: (string-append "get " filename)]
           [self wait-for-completion]))
    (define-message ftp (get: file to: target)
      (and [self send-command: (string-append "get " file " " target)]
           [self wait-for-completion]))
    ;; Cd to given directory
    (define-message ftp (cd: dir)
      (and [self send-command: (string-append "cd " dir)]
           [self wait-for-completion]))
    (lambda () [ftp clone])))

(provide "ftp")


