;;***************************************************************************
;; mapdir.scm
;; mapdir.scm,v 1.1.1.1 1995/02/27 15:16:07 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; AUTODOC MODULE
;; Browse a file system hierarchy.
;; This module provides a function to browse a full file system hierarchy.
;; It is quite usefull to write scripts.
;***************************************************************************

(require "system")
(require "regexp")

;; AUTODOC AUTOLOAD
;; (wildcard-to-regexp WILDEXPR)
;; Convert given (string) WILDEXPR to a regular expression.
;; Given a UNIX wildcard string, this function returns a regular expression
;; which will recognize the same set of strings.

(define (wildcard-to-regexp wild)
  (let ((e (string-length wild)))
    (let loop ((s 0) (re "^"))
      (if (< s e)
          (let ((c (string-ref wild s)))
            (case c
              ((#\\ #\. #\$) (loop (+ s 1) (string-append re (string #\\ c))))
              ((#\?) (loop (+ s 1) (string-append re ".")))
              ((#\*) (loop (+ s 1) (string-append re ".*")))
              (else (loop (+ s 1) (string-append re (string c))))))
          (regexp:make (string-append re "$"))))) )
      
;; AUTODOC AUTOLOAD
;; (mapdir root dirfct filefct fi fe di de)
;; Browse a file system, applying functions on every found file.
;; This function browse a file system starting a the given (string) ROOT. 
;; The DIRFCT function is applied to encoutered directories, it should return
;; #t if the directory is to be browsed recurseviley, and #f otherwise.
;; The FILEFCT function is applied with the full pathname of encoutered files.
;; FI should be a wildcard string indicating what files are to be included by
;; the search, FE is a wildcard string indicating files to be excluded. When
;; a file is encoutered, it is checked against the FI wildcard expr. If it 
;; matches, it is checked against the excluded wildcard expression, if this 
;; last match fails, the FILEFCT function is applied to it.
;; DI and DE are used for directories.

(define (mapdir root dirfct fct fi fe di de)
  (define (aux path files)
    (for-each (lambda (file) 
                (let ((fpath (path-append path file)))
                  ;; link typed files may be broken, this is detected by
                  ;; the file-exists call
                  (and (file-exists? fpath)
                       (if (directory? fpath)
                           (and dirfct 
                                (regexp:match-one file di)
                                (not (regexp:match-one file de))
                                (dirfct fpath)
                                (aux fpath (directory fpath)))
                           (and fct 
                                (regexp:match-one file fi)
                                (not (regexp:match-one file fe))
                                (fct fpath))))))
              files )
    #t )
  (set! fi (and fi (map wildcard-to-regexp fi)))
  (set! fe (and fe (map wildcard-to-regexp fe)))
  (set! di (and di (map wildcard-to-regexp di)))
  (set! de (and de (map wildcard-to-regexp de)))
  (set! root (expand-path-for-tilde root))
  (aux root (directory root)) )
  
(provide "mapdir")
