;;***************************************************************************
;; mosaic.scm
;; mosaic.scm,v 1.1.1.1 1995/02/27 15:16:07 baird Exp
;; See the file COPYRIGHT foe licensing conditions.
;; A simple scheme pilot for mosaic. Will only work if the script is run
;; on the same machine than Mosaic
;; AUTODOC MODULE
;; This module implements the Mosaic version 2.4 inter process communication.
;; It allows to control Mosaic browser through the following operations:
;; - opening a new URL
;; - creating a new window.
;***************************************************************************

(require "process")
(require "error")

;; AUTODOC AUTOLOAD
;; *mosaic:program*
;; Program to launch in order to start a new Mosaic process.

(define *mosaic:program* "Mosaic")

;; AUTODOC AUTOLOAD
;; *mosaic:pid-file*
;; The file of the name Mosaic uses to store its process id.

(define *mosaic:pid-file* (expand-path-for-tilde "~/.mosaicpid"))

;; AUTODOC AUTOLOAD
;; (mosaic:make-pilot)
;; Starts Mosaic, and make a controller for it.
;; This primitive builds an object to control Mosaic (typicaly to be used by
;; some helper applications). It uses the CCI interface of Mosaic 2.4.
;; It defines the following methods:
;;     . send-command: STRING: send STRING as a command to Mosaic,
;;     . goto: URL           : make Mosaic open given URL (a string),
;;     . newwin: URL         : make Mosaic opens URL in a new window.
;; When Mosaic is not running, the object tries to launch it (the first command
;; generally fails, because of the start up time of Mosaic, and the home URL).
;; Deathes of Mosaic are trapped wither through signal sending failure, or
;; changes in ~/.mosaicpid modify time.

(define mosaic:make-pilot
  (let ((pilot [obj:root clone]))
    [pilot defineSlot: "last-url" init: #f]
    [pilot defineSlot: "pid" init: #f]
    [pilot defineSlot: "last-checked" init: 0]
    ;; Re-read mosaic pid file (.mosaicpid file msut be checked before call):
    (define-message pilot (updatePid)
      [self pid: (with-input-from-file *mosaic:pid-file* (lambda () (read)))])
    ;; If autoStart is on, than launch a new Mosaic
    ;; Wait for Mosaic to write its mosaicpid file
    (define-message pilot (autoStart) 
      (let* ((pid-file *mosaic:pid-file*)
             (stat (and (file-exists? pid-file) (file-stat pid-file))))
        (if [self last-url]
            (make-process *mosaic:program* "-home" [self last-url])
            (make-process *mosaic:program*))
        (let loop ((nstat (and (file-exists? pid-file) (file-stat pid-file))))
            (sleep 1)
            (or (and (not stat) nstat)
                (and stat nstat (> [nstat mtime] [stat mtime]))
                (loop (and (file-exists? pid-file) (file-stat pid-file)))))
        ;; Mosaic is *really* broken :-(
        (sleep 2)))
    ;; First kill to mosaic has failed.
    ;; Check that our pid info is up to date (or auto-start), and re-kill
    (define-message pilot (checkAndKill)
      (let ((pid-file *mosaic:pid-file*))
            (let ((stat (and (file-exists? pid-file) (file-stat pid-file))))
              (if (and stat (> [stat mtime] [self last-checked]))
                  ;; Mosaic has been killed, and potentially re-run
                  [self updatePid]
                  ;; Mosaic has *never* been run, or is out dated
                  [self autoStart])
              [self last-checked: [stat mtime]]
              ;; We should never loop infintely: we will always end up by a 
              ;; call to autoStart...The sleep is needed here because of
              ;; Mosaic (as I understand)
              (or (error:case 
                   (begin (sleep 1) (kill-process [self pid] signal:sigusr1))
                   ((error:bad-syscall)
                    (lambda (e c . rest)
                      (and (string=? (list-ref c 0)  "kill") #f))))
                  [self checkAndKill]))) )
    ;; Send a command to Mosaic:
    (define-message pilot (send-command: cmd)
      (let ((cmd-file (string-append "/tmp/Mosaic."
                                     (integer->string [self pid]))))
        (with-output-to-file cmd-file (lambda () (display cmd)))
        (or (error:case (kill-process [self pid] signal:sigusr1)
                        ((error:bad-syscall)
                         (lambda (e ctxt . rest)
                           (and (string=? (list-ref ctxt 0)  "kill") #f))))
            [self checkAndKill])) )
    ;; Goodies:
    (define-message pilot (goto: url)
      [self last-url: url]
      [self send-command: (string-append "goto\n" url "\n")])
    (define-message pilot (newwin: url)
      [self last-url: url]
      [self send-command: (string-append "newwin\n" url "\n")])
    ;; Initialize:
    (if (file-exists? *mosaic:pid-file*)
        [pilot updatePid]
        [pilot autoStart])
  (lambda () pilot)) )

(provide "mosaic")


