/***************************************************************************
 * macro.c
 * macro.c,v 1.2 1995/03/17 14:22:53 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * OScheme macro expander.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

static OsAtom osa_mcons   = (OsAtom) 0 ;
static OsAtom osa_append  = (OsAtom) 0 ;
static OsAtom osa_lambda  = (OsAtom) 0 ;

static OsError ose_macro_ctxt = (OsError) 0 ;

/* We makee provision for three types of macro:
 * - C written macro:
 *    Which are to be installed through OsCExpanderInstall, the c
 *    expander takes as paremeter, the full form to be expanded. 
 * - Scheme call macro:
 *    Which are to be installed in scheme, with the define-macro form,
 *    they are called, as the C expander, with the full form to be
 *    expanded. This allows for more error checking
 * - Scheme 'normal' macro:
 *    Which are to be installed in scheme with the defmacro form.
 */

typedef struct _os_macro_expander {
    unsigned char type ;
    OsCodeBlock   block ;
} os_macro_expander_rec, *os_macro_expander ;

#define OS_C_MACRO	     	   1
#define OS_SCHEME_CALL_MACRO 	   2
#define OS_SCHEME_NORMAL_MACRO     3

/*
 * Managing macros space:
 */

static os_macro_expander make_macro_expander (OsInterp interp
                                              , OsAtom name
                                              , unsigned char type
                                              , OsCodeBlock block)
{
    os_macro_expander expander = OsGCMalloc (sizeof (os_macro_expander_rec)) ;
    expander->type  = type ;
    expander->block = block ;
    if ( OsHashtableGet (interp->macro_space, name) ) {
        OsOutputPortWriteCString (OsCurrentErrorPort
                                  , ";; Warning: redefining expander ") ;
        OsSendDisplay (name, OsCurrentErrorPort) ;
        OsOutputPortWriteChar (OsCurrentErrorPort, '\n') ;
    }
    OsHashtablePut (interp->macro_space, name, expander) ;
    return (expander) ;
}

OsObject OsCExpanderInstall (OsInterp interp
                             , char *cname
                             , OsObject (*cexpd)()) {
    OsAtom      atom  = OsAtomMake (interp, cname, OsA_Unbound) ;
    OsCodeBlock block = OsCodeBlockMakeForC (cexpd, 1, OS_SECURITY_LOWEST) ;

    (void) make_macro_expander (interp, atom, OS_C_MACRO, block) ;
    return ((OsObject) atom) ;
}

OsObject OsSCallExpanderInstall (OsInterp interp
                                 , OsAtom name
                                 , OsCodeBlock block) {
    (void) make_macro_expander (interp, name, OS_SCHEME_CALL_MACRO, block) ;
    return ((OsObject) name) ;
}

OsObject OsSNormalExpanderInstall (OsInterp interp
                                   , OsAtom name
                                   , OsCodeBlock block) {
    (void) make_macro_expander (interp, name, OS_SCHEME_NORMAL_MACRO, block) ;
    return ((OsObject) name) ;
}
    
/*
 * The expander proper (small isn't it !)
 * Entry point is OsExpandExpr
 */

static OsObject os_expd_args (OsObject args) {
    OsObject start = OsA_Nil, *list = &start ;

    if ( ! OsIsCons (args) ) {
        return (OsExpandExpr (args)) ;
    }
    do {
        *list = (OsObject) OsConsMake(OsExpandExpr(OS_CONS_CAR(args)),OsA_Nil);
        list  = &(OS_CONS_CDR (*list)) ;
        args  = OS_CONS_CDR (args) ;
    } while ( OsIsCons (args) ) ;
    return (start) ;
}

static OsObject os_expd_cons (OsCons call) {
    OsInterp interp = OsCurrentInterp() ;
    OsObject fun  = OS_CONS_CAR (call) ;
    OsObject args = OS_CONS_CDR (call) ;
    os_macro_expander expd ;

    OsCodeBlock block = (OsCodeBlock) OsHashtableGet (interp->macro_space
                                                      , fun) ;
    if (expd = (os_macro_expander) OsHashtableGet (interp->macro_space, fun)) {
        OsCodeBlock blk ;
        OsObject    stage, error ;
        switch (expd->type) {
          case OS_C_MACRO:
          case OS_SCHEME_CALL_MACRO:
            stage = OsCodeBlockApplyWithArray (interp
                                               , expd->block
                                               , 1, (OsObject *) &call
                                               , &error) ;
            break ;
          case OS_SCHEME_NORMAL_MACRO:
            stage = OsCodeBlockApplyWithList (interp
                                              , expd->block
                                              , (OsCons) args
                                              , &error) ;
            break ;
          default:
            OsErrorTrigger (1, OsE_Implementation
                            , OsStringMake ("os_expd_cons")
                            , OsStringMake ("invalid expander type.")) ;
            /* NOT REACHED */
        }
        if ( error ) {
            os_uncatched_error (OsErrorAugment (error, 1, ose_macro_ctxt
                                                , fun, call));
        } else {
            if ( (stage != (OsObject) call) )
                return (OsExpandExpr (stage)) ;
        }
    } 
    if ( fun == (OsObject) osa_lambda ) {   /* don't expd lambda formals */
        OsMustBeCons(args) ;
        OsMustBeCons (OS_CONS_CDR (args)) ;
        return ((OsObject) OsList2Append (fun
                                          , OS_CONS_CAR (args)
                                          , os_expd_args(OS_CONS_CDR(args))));
    } else if ( fun != OsA_Quote ) {	    /* don't expd quoted exprs */
        return ((OsObject) OsConsMake (OsExpandExpr(fun),os_expd_args (args)));
    } else {
        return ((OsObject) call) ;
    }
    /* NOT REACHED */
}

OsObject OsExpandExpr (OsObject expr) {
    if ( OsIsCons (expr) ) {
        return (os_expd_cons ((OsCons) expr)) ;
    } else {
        return (expr) ;
    }
    /* NOTE REACHED */
}

/*
 * Specific predefined expander
 */

/* AUTODOC
 * (define-macro (NAME CALL) BODY)
 * Define a new syntactic macro, taking the full form as parameters.
 * This primitives defines a new macro named NAME. When a corresponding call
 * is performed, the full CALL is passed as a parameter to BODY, and the result
 * is considered to be the expansion of the macro call.
 */

static OsObject os_expd_define_macro (OsObject expr) {
    OsCons header, body, defun ;
    OsAtom name ;
    OsObject expfun ;
    OsCodeBlock block ;
    
    OsMustBeCons (expr) ;
    expr = OS_CONS_CDR (expr) ;
    OsMustBeCons (expr) ;
    header = (OsCons) OS_CONS_CAR (expr) ;
    body   = (OsCons) OS_CONS_CDR (expr) ;
    OsMustBeCons (body) ;
    name = (OsAtom) OS_CONS_CAR (header) ;
    OsMustBeAtom (name) ;
    header = (OsCons) OS_CONS_CDR (header) ;
    defun  = OsList2Append (OsA_Lambda, (OsObject) header, (OsObject) body) ;
    expfun = OsExpandExpr ((OsObject) defun) ;
    block  = OsCompileExpr (expfun) ;
    (void) OsSCallExpanderInstall (OsCurrentInterp()
                                   , name
                                   , (OsCodeBlock) OsCodeBlockExecute (OsCurrentInterp(), block));
    return (OsA_True) ;
}

/* AUTODOC
 * (defmacro (NAME ARG*) BODY)
 * Define a new syntactic macro, with parameters.
 * Define a new syntactic macro named NAME. The BODY of the macro is executed
 * in an environments binding all ARG to the arguments to the macro call.
 */
 
static OsObject os_expd_defmacro (OsObject expr) {
    OsCons      header, body, formals, defun ;
    OsAtom      name ;
    OsObject    expfun ;
    OsCodeBlock block ;
    
    OsMustBeCons (expr) ;
    expr = OS_CONS_CDR (expr) ;
    OsMustBeCons (expr) ;
    header = (OsCons) OS_CONS_CAR (expr) ;
    OsShouldBeCons (header) ;
    name    = (OsAtom) OS_CONS_CAR (header) ;
    formals = (OsCons) OS_CONS_CDR (header) ;
    body    = (OsCons) OS_CONS_CDR (expr) ;
    OsMustBeCons (body) ;
    defun  = OsList2Append (OsA_Lambda, (OsObject) formals, (OsObject) body) ;
    expfun = OsExpandExpr ((OsObject) defun) ;
    block  = OsCompileExpr (expfun) ;
    (void) OsSNormalExpanderInstall (OsCurrentInterp()
                                     , name
                                     , (OsCodeBlock)OsCodeBlockExecute(OsCurrentInterp(), block));
    return (OsA_True) ;
}

static OsObject os_expd_quasiquote_cons (OsObject expr) {
    OsObject start, *list ;
    OsObject item ;
    OsObject tmp = (OsObject) 0 ;
    
    start = (OsObject) OsConsMake ((OsObject) osa_mcons, OsA_Nil) ;
    list  = &(OS_CONS_CDR (start)) ;
    do {
        item = OS_CONS_CAR (expr) ;
        if ( OsIsCons (item) ) {
            OsObject fun = OS_CONS_CAR (item) ;
            if ( fun == OsA_Unquote ) {
                if ( tmp ) {
                    *list = (OsObject) OsConsMake (tmp, OsA_Nil) ;
                    list  = &(OS_CONS_CDR (tmp)) ;
                    tmp   = (OsObject) 0 ;
                }
                *list = (OsObject) OsConsMake (OS_CONS_CAR (OS_CONS_CDR(item))
                                               , OsA_Nil) ;
                list  = &(OS_CONS_CDR (*list)) ;
            } else if ( fun == OsA_UnquoteSplice ) {
                OsObject unquoted ;
                OsMustBeCons (OS_CONS_CDR (item)) ;
                *list = (OsObject) OsConsMake (OsA_Nil, OsA_Nil) ;
                unquoted = OS_CONS_CAR (OS_CONS_CDR (item)) ;
                tmp   = (OsObject) OsConsMake (unquoted, OsA_Nil) ;
                list  = &(OS_CONS_CDR (tmp)) ;
                start = (OsObject) OsList2Append ((OsObject) osa_append
                                                  , start
                                                  , tmp) ;
                tmp   = (OsObject) OsConsMake ((OsObject) osa_mcons, OsA_Nil);
            } else {
                OsObject unquoted = os_expd_quasiquote_cons (item) ;
                if ( tmp ) {
                    *list = (OsObject) OsConsMake (tmp, OsA_Nil) ;
                    list  = &(OS_CONS_CDR (tmp)) ;
                    tmp   = (OsObject) 0 ;
                }
                *list = (OsObject) OsConsMake (unquoted, OsA_Nil) ;
                list  = &(OS_CONS_CDR (*list)) ;
            }
        } else if (((item == OsA_Unquote) || (item == OsA_UnquoteSplice))
                   && (OsIsCons (OS_CONS_CDR(expr)))
                   && (OS_CONS_CDR (OS_CONS_CDR (expr)) == OsA_Nil)) {
            OsObject obj = OS_CONS_CAR (OS_CONS_CDR(expr)) ;
            /* At end of template, with a non nil cdr */
            if ( item == OsA_Unquote ) {
                *list = (OsObject) OsConsMake (obj, OsA_Nil) ;
            } else /* unquote-splice */ {
                *list = obj ;
            }
            return (start) ;
        } else {
            OsObject quoted ;	/* I HATE > 80 cols line ;-) */
            if ( tmp ) {
                *list = (OsObject) OsConsMake (tmp, OsA_Nil) ;
                list  = &(OS_CONS_CDR (tmp)) ;
                tmp   = (OsObject) 0 ;
            }
            quoted = (OsObject) OsList2Make (OsA_Quote, item) ;
            *list  = (OsObject) OsConsMake (quoted, OsA_Nil);
            list   = &(OS_CONS_CDR (*list)) ;
        }
    } while ( expr = OS_CONS_CDR (expr), OsIsCons (expr) ) ;
    *list = (OsObject) OsConsMake ((OsObject) OsList2Make (OsA_Quote, expr)
                                   , OsA_Nil) ;
    /* OsConsMake (expr, OsA_Nil) ;*/
    return (start) ;
    /* NOT REACHED */
}

static OsObject os_expd_quasiquote_vector (OsObject expr) { /* POOR ! */
    OsObject     start = OsA_Nil, *plist = &start ;
    OsObject     exp ;
    OsVector     vector = OsVectorMake (0) ;
    unsigned int idx = 0, len = OsVectorLength (expr) ;

    /* vector->list */
    while ( idx < len ) {
        *plist = (OsObject) OsConsMake (OsVectorRef (expr, idx), OsA_Nil) ;
        plist  = &(OsConsCdr (*plist)) ;
        idx++ ;
    }
    exp = os_expd_quasiquote_cons (start) ;
    return ((OsObject) OsList2Make ((OsObject) OsAtomMake (OsCurrentInterp()
                                                           , "list->vector"
                                                           , OsA_Unbound)
                                    , exp)) ;
}

static OsObject os_expd_quasiquote (OsObject expr) {
    OsMustBeCons (expr) ;
    expr = OS_CONS_CDR(expr) ;
    OsMustBeCons (expr) ;
    expr = OS_CONS_CAR (expr) ;
    if ( OsIsCons (expr) ) {
        return (os_expd_quasiquote_cons (expr)) ;
    } else if ( OsIsVector (expr) ) {
        return (os_expd_quasiquote_vector (expr)) ;
    } else {
        return ((OsObject) OsList2Make (OsA_Quote, expr)) ;
    }
    /* NOT REACHED */
}

/*
 * Subrs:
 */

/* AUTODOC
 * (sys:expand EXPRESSION)
 * Expand an expression with the currently defined set of macros.
 * This function expand the given EXPRESSION until it contains no more macros
 * call. It returns the expanded EXPRESSION.
 */
 
static OsObject OsSubrExpand (obj) 
    OsObject obj ;
{
    return (OsExpandExpr (obj)) ;
}

/*
 * Initializer:
 */
    
void OsExpanderInitialize (OsInterp interp, int *pargc, char **argv) {
    ose_macro_ctxt = OsErrorMake (interp
                                  , "error:macro-ctxt"
                                  , "macro %0 with %1 form"
                                  , 2, OSE_CATCHABLE) ;
    osa_mcons  = OsAtomMake (interp, "cons*", OsA_Unbound) ;
    osa_append = OsAtomMake (interp, "append", OsA_Unbound) ;
    osa_lambda = OsAtomMake (interp, "lambda", OsA_Unbound) ;
    return ;
}

void OsExpanderSubrInitialize (OsInterp interp) {
    OsCExpanderInstall (interp, "define-macro", os_expd_define_macro) ;
    OsCExpanderInstall (interp, "defmacro", os_expd_defmacro) ;
    OsCExpanderInstall (interp, "quasiquote", os_expd_quasiquote) ;
    OsSubrMake (interp, "sys:expand", OsSubrExpand, 1) ;
}

