/***************************************************************************
 * os_regexp.c
 * os_regexp.c,v 1.1.1.1 1995/02/27 15:15:50 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 *
 * AUTODOC MODULE
 * This module provides an scheme interface to H. Spencer regexp package. It
 * allows to search for regular expressions in strings, and to retrieve part
 * of matching strings.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

OsClass OsRegexpClass = (OsClass) 0 ;

static OsError ose_regexp_error = (OsError) 0 ;

/*
 * Constructors:
 */

OsRegexp OsRegexpMake (unsigned char *source) {
    OsRegexp re = (OsRegexp) OsObjectMake (OsRegexpClass, 0) ;
    re->prog   = os_regcomp (source) ;
    re->source = OsStringMake (source) ;
    return (re) ;
}

OsRegexp OsRegexpMakeWithString (OsString source) {
    OsRegexp re = (OsRegexp) OsObjectMake (OsRegexpClass, 0) ;
    re->prog   = os_regcomp (source->string) ;
    re->source = source ;
    return (re) ;
}

/*
 * regexp package interface:
 */

void os_regerror (char *msg) {
    (void) OsErrorTrigger (1, ose_regexp_error, OsStringMake (msg)) ;
}

static char *os_regsub (prog, source)
    os_regexp *prog ;
    char *source ;
{
    char *src;
    char *dst;
    char c;
    int no;
    int len;
    char *dest = (char *) OsGCMallocAtomic (12);
    char *lastdest = dest+11;		/* last allowable - 1 */

    if (prog == NULL || source == NULL) {
	os_regerror("NULL parm to regsub");
	/*NOTREACHED*/
    }
    if (UCHARAT(prog->program) != OL_MAGIC) {
	os_regerror("damaged regexp fed to regsub");
	/*NOTREACHED*/
    }

    src = source;
    dst = dest;
    while ((c = *src++) != '\0') {
	if (c == '&')
	    no = 0;
	else if (c == '\\' && '0' <= *src && *src <= '9')
	    no = *src++ - '0';
	else
	    no = -1;

	if (no < 0) {			/* Ordinary character. */
	    if (dst + 1 >= lastdest) {
		char *olddest = dest;
		int newlength = ((lastdest - dest) + 5) * 2 - 4;
		dest = (char *) OsGCRealloc(dest, newlength);
		lastdest = dest + (newlength - 1);
		dst = dest + (dst - olddest);
	    }
	    *dst++ = c;
	} else if (prog->startp[no] != NULL && prog->endp[no] != NULL) {
	    len = prog->endp[no] - prog->startp[no];
	    if (dst + len >= lastdest) {
		char *olddest = dest;
		int newlength = ((lastdest - dest) + 5) * 2 - 4;

		while (newlength <= (dst - dest) + len)
		    newlength = ((newlength) + 5) * 2 - 4;
		dest = (char *) OsGCRealloc(dest, newlength);
		lastdest = dest + (newlength - 1);
		dst = dest + (dst - olddest);
	    }
	    
	    (void) strncpy(dst, prog->startp[no], len);
	    dst += len;
	    if (dst > dest && *(dst-1) == '\0') {	/* strncpy hit NUL. */
		os_regerror("damaged match string");
		/*NOTREACHED*/
	    }
	}
    }
    *dst++ = '\0';
    return (dest) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (regexp? OBJECT)
 * Is OBJECT a compiled regular expression ?
 */

static OsObject OsSubrRegexpP (OsObject object) {
    return (OsIsRegexp (object) ? OsA_True : OsA_False) ;
}

/* AUTODOC
 * (regexp:make STRING)
 * Make a regular expression out of the given string.
 * This primitive compiles the STRING as a regular expression, The STRING 
 * should respect the standard regular expression syntax, augmented with the
 * facility to group part of the matching parts by using parenthesis around it.
 * For example:
 * (define re (regexp:make "^([^ ]*) (.*)$"))
 * Will parse a line containg two strings separated by a white space, and 
 * will be able to return both string.
 */

static OsObject OsSubrRegexpMake (OsString source) {
    OsMustBeString (source) ;
    return ((OsObject) OsRegexpMakeWithString (source)) ;
}

/* AUTODOC
 * (regexp:match REGEXP STRING [PARTS*])
 * Match REGEXP against STRING, and return a list of subparts.
 * This primitive matches the REGEXP against the STRING. If no PARTS* are 
 * provided it returns a boolean indicating whether the match succeeded.
 * PARTS* should be a list of integer indicating which subparts you want to
 * retreive (see 'regexp:make). Part 0 is the whole string. If PARTS* are
 * provided than the match returns the list of sub parts indicated, otherwise
 * it returns #f.
 */

static OsObject OsSubrRegexpMatch (unsigned int argc, OsObject *argv) {
    OsRegexp re ;
    OsString pat, sub ;
    if ( argc < 2 ) {
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (2)
                               , OsIntegerMake (argc)) ;
    }
    re  = (OsRegexp) OsMustBeRegexp (argv[0]) ;
    pat = (OsString) OsMustBeString (argv[1]) ;
    argc -= 2 ; argv+= 2 ;
    if ( os_regexec (re->prog, pat->string) ) {
        os_regexp *prog = re->prog ;
        OsObject  start = OsA_True, *list = &start ;
        int       i ;
        while ( argc-- ) {
            OsMustBeInteger (*argv) ;
            i = ((OsInteger) *argv)->ival ;
            if ((i>=0) && (i<NSUBEXP) && (prog->startp[i] || prog->endp[i])) {
                sub = OsStringMakeFromBytes (prog->endp[i] - prog->startp[i]
                                             , prog->startp[i]) ;
            } else {
                sub = OsStringMake ("") ;
            }
            *list = (OsObject) OsConsMake ((OsObject) sub, OsA_Nil) ;
            list  = &(OS_CONS_CDR (*list)) ;
            argv++ ;
        }
        return (start) ;
    }
    return (OsA_False) ;
}

/* AUTODOC
 * (regexp:sub REGEXP STRING PATTERN)
 * Match REGEXP against STRING and format sub parts into PATTERN.
 * This function matches REGEXP against STRING. If the match fails, it returns
 * #f, otherwise, it replaces all occurences of '\n' (where n is a single 
 * digit) with the corresponding matching subpart, and returns the resulting
 * string.
 */

static OsObject OsSubrRegexpSub (OsRegexp re, OsString pat, OsString sub) {
    os_regexp *prog ;
    OsMustBeRegexp (re) ;
    OsMustBeString (pat) ;
    OsMustBeString (sub) ;
    if ( ! (os_regexec (prog = re->prog, pat->string)) )
        return (OsA_False) ;
    else
        return ((OsObject) OsStringMake (os_regsub (prog, sub->string))) ;
}

/*
 * Message implementations:
 */

static OsObject os_regexp_display (OsRegexp re, int sel, OsOutputPort port) {
    OsSendDisplay (re->source, port) ;
    return ((OsObject) re) ;
}

static OsObject os_regexp_equal (OsRegexp self, int sel, OsRegexp other) {
    return (OsSendEqual (self->source, other->source)) ;
}

static OsObject os_regexp_finalize (OsRegexp self) {
    free (self->prog) ;
    return (OsA_True) ;
}

/*
 * Initialize :
 */

void OsRegexpInitialize (OsInterp interp, int *pargc, char **argv) {
    OsRegexpClass = OsClassDeclare ("Regexp", sizeof (OsRegexpRec)) ;
    OsDefineMsg (OsRegexpClass, OS_SEL_DISPLAY, os_regexp_display) ;
    OsDefineMsg (OsRegexpClass, OS_SEL_EQUAL, os_regexp_equal) ;
    OsDefineMsg (OsRegexpClass, OS_SEL_FINALIZE, os_regexp_finalize) ;
    return ;
}

void OsRegexpSubrInitialize (OsInterp interp) {
    ose_regexp_error = OsErrorMake (interp
                                    , "error:regexp-error"
                                    , "regexp error %0"
                                    , 1, OSE_SEVERE) ;
    OsSubrMake (interp, "regexp?", OsSubrRegexpP, 1) ;
    OsSubrMake (interp, "regexp:make", OsSubrRegexpMake, 1) ;
    OsSubrMake (interp, "regexp:match", OsSubrRegexpMatch, OS_NARY) ;
    OsSubrMake (interp, "regexp:sub", OsSubrRegexpSub, 3) ;
    return ;
}
