/***************************************************************************
 * sig.c
 * sig.c,v 1.2 1995/03/12 02:58:34 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * This modules provides synchronous signal management, on top of the raw
 * asynchronous one: signal handlers (be them C or scheme) are called only
 * when the VM state makes it possible.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>

extern int errno ;

#include <oscheme.h>

#ifdef OS_HAS_SIGNALS

int os_signals_pending = 0 ;

typedef struct _os_handler {
    OsClosedBlock block ;
    void          (*cfun)() ;
} os_handler_rec, *os_handler ;

static os_handler   os_handlers    = (os_handler) 0 ;
static unsigned int os_handlers_sz = 0 ;

/*
 * Internally used functions:
 */

static int *pendings = (int *) 0 ;
static int *last     = (int *) 0 ;

static void os_default_signal_handler (signo)
    int signo ;
{
/*
 * Queue the signal for future delivery to scheme:
 */
    if ( ( ! last)
        || (last + 1 >= pendings + OS_SIGNALS_MAX_PENDING) ) {
        fprintf (stderr, "os_default_signal_handler: loosing signal %d\n"
                 , signo) ;
    } else {
        *last = signo ;
        last++ ;
        os_signals_pending = 1 ;
    }
/*
 * Restart it if needed:
 */
#ifdef OS_REDECLARE_SIGNALS
    signal (signo, os_default_signal_handler) ;
#endif /* OS_REDECLARE_SIGNALS */
    return ;
}

static void register_sighandler (signo, cfun, block)
    int           signo ;
    void          (*cfun)() ;
    OsClosedBlock block ;
{
    if ( signo >= os_handlers_sz ) {
        unsigned int sz = (signo + 1) * sizeof (os_handler_rec) ;
        if ( os_handlers_sz ) {
            os_handlers = (os_handler) OsGCRealloc (os_handlers, sz) ;
        } else {
            os_handlers = (os_handler) OsGCMallocAtomic (sz) ;
        }
        os_handlers_sz = signo + 1 ;
    }
    os_handlers[signo].block = block ;
    os_handlers[signo].cfun  = cfun ;
    return ;
}

/*
 * C-API:
 */

void OsDoSignals () {
    OsClosedBlock block ;
    void          (*cfun)() ;

    if ( os_signals_pending )
        os_signals_pending = 0 ;
    else
        return ;
    while (--last >= pendings) {
        if ( *last < os_handlers_sz ) {
            if ( (cfun = os_handlers[*last].cfun) ) {
                (*cfun)(*last) ;
            } else if ( block = os_handlers[*last].block ) {
                OsObject ival = (OsObject) OsIntegerMake (*last) ;
                (void) OsCodeBlockApplyWithArray (OsCurrentInterp()
                                                  , (OsCodeBlock) block
                                                  , 1, (OsObject *) &last
                                                  , (OsObject *) 0) ;
            }
        }
    }
    last = pendings ;
    return ;
}

void OsCSignalHandle (signo, cfun)
    int  signo ;
    void (*cfun)() ;
{
#ifdef SIG_ERR
    if ( signal (signo, os_default_signal_handler) == SIG_ERR ) 
#else
    if ( signal (signo, os_default_signal_handler) < 0 ) 
#endif
    {
        (void) OsErrorTrigger (1, OsE_Implementation
                               , OsStringMake ("signal")
                               , OsStringMake ("invalid signal syscall.")) ;
    }
    register_sighandler (signo, cfun, (OsClosedBlock) 0) ;
    return ;
}
    
/*
 * Subrs:
 */

/* AUTODOC 
 * (signal SIGNO THUNK)
 * Catch signal SIGNO, and invoke THUNK when it is triggered.
 * This primitive allows to catch signal SIGNO. When this signal is sent to the
 * oscheme process, the provided THUNK is applied to the signal number which
 * triggered it.
 * This primitive triggers the 'error:bad-syscall if the underlying "signal"
 * system call fails.
 * Signal are delivered to scheme code in a synchornous fashion, hence there
 * no restriction is applied to THUNK (i.e: it can allocate objects, etc).
 * WARNING: oscheme uses a stack to record signals before they are delivered
 * to the scheme layer. When this stack is exceeded (in size), signals are
 * just lost (a warning message is displayed when such situations occur).
 */
   
static OsObject OsSubrSignal (signo, handler)
    OsInteger     signo ;
    OsClosedBlock handler ;
{
    OsShouldBeInteger (signo) ;
    OsShouldBeClosedBlock (handler) ;

#ifdef SIG_ERR
    if ( signal (signo->ival, os_default_signal_handler) == SIG_ERR )
#else
    if ( signal (signo->ival, os_default_signal_handler) < 0 )
#endif
    {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("signal")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) signo->ival, OsA_Nil));
    } else {
        register_sighandler (signo->ival, (void (*)()) 0, handler) ;
    }
    return ((OsObject) signo) ;
}
                
/*
 * Initialization:
 */

void OsSignalsInitialize (OsInterp interp, int *pargc, char **argv) {
    pendings = (int *) OsGCMallocAtomic (sizeof(int) * OS_SIGNALS_MAX_PENDING);
    last     = pendings ;
    return ;
}

void OsSignalsSubrInitialize (OsInterp interp) {
    OsSecSubrMake (interp, "signal", OsSubrSignal
                   , 2, OS_SECURITY_LEVEL_A) ;
    OsPathProvides (interp, OsStringMake ("signal")) ;
    OsAtomMake (interp, "signal:sighup", (OsObject) OsIntegerMake (SIGHUP)) ;
    OsAtomMake (interp, "signal:sigint", (OsObject) OsIntegerMake (SIGINT)) ;
    OsAtomMake (interp, "signal:sigquit", (OsObject) OsIntegerMake (SIGQUIT)) ;
    OsAtomMake (interp, "signal:sigusr1", (OsObject) OsIntegerMake (SIGUSR1)) ;
    OsAtomMake (interp, "signal:sigusr2", (OsObject) OsIntegerMake (SIGUSR2)) ;
    OsAtomMake (interp, "signal:sigpipe", (OsObject) OsIntegerMake (SIGPIPE)) ;
    OsAtomMake (interp, "signal:sigchld", (OsObject) OsIntegerMake (SIGCHLD)) ;
    return ;
}

#endif /* OS_HAS_SIGNALS */
