/***************************************************************************
 * sys.c
 * sys.c,v 1.1.1.1 1995/02/27 15:15:46 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * Required system interface.
 * This module provides the required oscheme interface to the underlying
 * system. Additional system facilities (depending on the system) are
 * available as dynamic modules.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <stdlib.h>
#ifdef OS_HAS_UNISTD_H
#include <unistd.h>
#endif
#include <pwd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#ifdef OS_HAS_SYSDIR_H
#    include <sys/dir.h>
#    define dirent direct
#else
#    include <dirent.h>
#endif
#include <sys/stat.h>
#ifdef OS_HAS_SYSFILE_H
#    include <sys/file.h>
#endif
#include <sys/param.h>
#include <errno.h>

extern int errno ;

#include <oscheme.h>

/*
 * Environment management:
 */

static OsHashtable os_env = (OsHashtable) 0 ;

static void os_env_initialize () {
    extern char  **environ ;
    char         **penv = environ, *var, *val ;
    unsigned int varlen = 0 ;
    
    os_env = OsHashtableMake (0
                              , (OsHashFct) OsStringHash
                              , (OsCmpFct) OsStringCompare) ;
    while ( *penv ) {
        for (val = var = *penv, varlen = 0 ; *val && (*val != '=') ; val++ )
            varlen++ ;
        if ( *val == '=' ) {
            OsHashtablePut (os_env
                            , OsStringMakeFromBytes (varlen, var)
                            , OsStringMake (val+1)) ;
        }
        penv++ ;
    }
    return ;
}

OsString OsGetenv (string, def)
    OsString string, def ;
{
    OsString val = (OsString) OsHashtableGet (os_env, string) ;
    return (val ? val : def) ;
}

static void env_append (var, val, vector)
    OsString var, val ;
    OsVector vector ;
{
    char *line ;
    line = (char *) OsGCMallocAtomic (var->len+val->len+3) ;
    sprintf (line, "%s=%s", var->string, val->string) ;
    (void) OsVectorFastAppend (vector, (OsObject) line) ;
    return ;
}
    
void OsEnvUpdate ()
{
    extern char **environ ;
    OsVector envv = OsVectorMake (0) ;
    OsHashtableMap (os_env, env_append, envv) ;
    (void) OsVectorFastAppend (envv, (OsObject) 0) ;
    environ = (char **) envv->cells ; 
    return ;
}

/* AUTODOC
 * (getenv VARIABLE)
 * Returns the binding of VARIABLE in process environment.
 * This primtive returns the string VARIABLE is bound to in the current
 * process environment. If the variable is not set, it returns #f.
 */

static OsObject OsSubrGetenv (OsString var) {
    OsString val ;
    OsMustBeString (var) ;
    if ( (val = (OsString) OsHashtableGet (os_env, var)) )
        return ((OsObject) val) ;
    else
        return (OsA_False) ;
}

/* AUTODOC
 * (putenv VARIABLE VALUE)
 * Set or change VARIABLE value to VALUE in process environment.
 * This function set or changes the current binding of VARIABLE in the
 * enviornment. If value if #f, than the VARIABLE is unset, otherwise, VALUE
 * should be a string describing the new binding.
 * All environment variables are passed to forked processes.
 */
   
static OsObject OsSubrPutenv (OsString var, OsString val) {
    OsMustBeString (var) ;
    if ( val == (OsString) OsA_False ) {
        OsHashtableRemove (os_env, var) ;
    } else {
        OsMustBeString (val) ;
        OsHashtablePut (os_env, var, val) ;
    }
    return ((OsObject) var) ;
}

/*
 * File manipulations
 */

static OsEObj stat_sclass = (OsEObj) 0 ;


static void stat_sclass_initialize () {
#define _SFSIZE(_f)	OsFieldSize(struct stat, _f)
    stat_sclass = OsStructClassMake ("Stat", sizeof (struct stat)) ;
    OsStructRSlotDeclare (stat_sclass
                          , "dev"
                          , OsStructAccessorFind (_SFSIZE (st_dev), 0)
                          , OsOffsetOf (struct stat, st_dev)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "ino"
                          , OsStructAccessorFind (_SFSIZE (st_ino), 0) 
                          , OsOffsetOf (struct stat, st_ino)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "mode"
                          , OsStructAccessorFind (_SFSIZE (st_mode), 0) 
                          , OsOffsetOf (struct stat, st_mode)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "uid"
                          , OsStructAccessorFind (_SFSIZE (st_uid), 0) 
                          , OsOffsetOf (struct stat, st_uid)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "gid"
                          , OsStructAccessorFind (_SFSIZE (st_gid), 0) 
                          , OsOffsetOf (struct stat, st_gid)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "size"
                          , OsStructAccessorFind (_SFSIZE (st_size), 0)
                          , OsOffsetOf (struct stat, st_size)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "atime"
                          , OsStructAccessorFind (_SFSIZE (st_atime), 0)
                          , OsOffsetOf (struct stat, st_atime)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "mtime"
                          , OsStructAccessorFind (_SFSIZE (st_mtime), 0)
                          , OsOffsetOf (struct stat, st_mtime)
                          , (void *) 0) ;
    OsStructRSlotDeclare (stat_sclass
                          , "ctime"
                          , OsStructAccessorFind (_SFSIZE (st_ctime), 0)
                          , OsOffsetOf (struct stat, st_ctime)
                          , (void *) 0) ;
    return ;
#undef _SFSIZE 
}

/* AUTODOC
 * (file-stat FILENAME)
 * Gather FILENAME informations.
 * This primitive returns all available informations on FILENAME, or triggers
 * an 'error:bad-syscall if the underlying "stat" system call fails.
 * The returned object has the following slots:
 *    dev   : device number,
 *    ino   : inode number,
 *    mode  : mode,
 *    uid   : user id of owner,
 *    gid   : group id of owner,
 *    size  : size of file in bytes,
 *    atime : last access time,
 *    mtime : last modification time,
 *    ctime : creation time.
 * All those slots are read-only.
 */

static OsObject OsSubrFileStat (OsString file) {
    struct stat *buf = OsGCMallocAtomic (sizeof (struct stat)) ;
    OsShouldBeString (file) ;
    if ( stat (file->string, buf) ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("stat")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) file, OsA_Nil)) ;
    } else {
        return ((OsObject) OsStructMake (stat_sclass, OsPointerMake (buf))) ;
    }
    /* NOT REACHED */
}

/* AUTODOC
 * (directory DIR)
 * Returns the content of given directory DIR.
 * This primitives returns the list of files directory DIR contains. It never
 * returns the current directory and its parent (even if available). It
 * triggers the 'error:bad-syscall if underlying system call "opendir" fails.
 */

static OsObject OsSubrDirectory (OsString path) {
    DIR           *dir ;
    struct dirent *entry ;
    OsObject      start = OsA_Nil, *list = &start ;
    
    OsMustBeString (path) ;
    if ( ! (dir = opendir (path->string)) ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("opendir")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) path, OsA_Nil)) ;
    }
    while ( (entry = readdir (dir)) ) {
        if (! strcmp (entry->d_name, ".") || ! strcmp (entry->d_name, ".."))
            continue ;
        *list = (OsObject) OsConsMake ((OsObject) OsStringMake (entry->d_name)
                                       , OsA_Nil) ;
        list  = &(OS_CONS_CDR (*list)) ;
    }
    closedir (dir) ;
    return (start) ;
}

OsObject OsFileExists (filename)
    char *filename ;
{
    return (access (filename, F_OK) ? OsA_False : OsA_True) ;
}

/* AUTODOC
 * (file-exists? FILENAME)
 * Does FILENAME exists ?
 * This function returns #t if the file exists, and #f otherwise.
 */

static OsObject OsSubrFileExistsP (filename)
    OsString filename ;
{
    OsShouldBeString (filename) ;
    return (OsFileExists (filename->string)) ;
}

/* AUTODOC
 * (exit STATUS)
 * Exit from oscheme, returning the given STATUS.
 * This primitives quits oscheme, and return as its exit code the integer
 * STATUS.
 */
 
static OsObject OsSubrExit (OsInteger code) {
    OsShouldBeInteger (code) ;
    exit (code->ival) ;
}

/*
 * Password entries
 */

static OsEObj pwd_sclass = (OsEObj) 0 ;

static void pwd_sclass_initialize () {
#define _SFSIZE(_f)	OsFieldSize (struct passwd, _f)
    pwd_sclass = OsStructClassMake ("Passwd", sizeof (struct passwd)) ;
    OsStructRSlotDeclare (pwd_sclass
                          , "name"
                          , OsStructAccessorString
                          , OsOffsetOf (struct passwd, pw_name)
                          , (void *) 0) ;
    OsStructRSlotDeclare (pwd_sclass
                          , "uid"
                          , OsStructAccessorFind (_SFSIZE(pw_uid), 0) 
                          , OsOffsetOf (struct passwd, pw_uid)
                          , (void *) 0) ;
    OsStructRSlotDeclare (pwd_sclass
                          , "gid"
                          , OsStructAccessorFind (_SFSIZE(pw_gid), 0) 
                          , OsOffsetOf (struct passwd, pw_gid)
                          , (void *) 0) ;
    OsStructRSlotDeclare (pwd_sclass
                          , "gecos"
                          , OsStructAccessorString
                          , OsOffsetOf (struct passwd, pw_gecos)
                          , (void *) 0) ;
    OsStructRSlotDeclare (pwd_sclass
                          , "dir"
                          , OsStructAccessorString
                          , OsOffsetOf (struct passwd, pw_dir)
                          , (void *) 0) ;
    OsStructRSlotDeclare (pwd_sclass
                          , "shell"
                          , OsStructAccessorString
                          , OsOffsetOf (struct passwd, pw_shell)
                          , (void *) 0) ;
#undef _SFSIZE
    return ;
}

OsString OsGetHomeDirectory (user)
    OsString user ;
{
    struct passwd *entry ;
    if ( ! user->len )
        return (OsGetenv (OsStringMake ("HOME"), (OsString) 0)) ;
    if ( (entry = getpwnam (user->string)) ) {
        return (OsStringMake (entry->pw_dir)) ;
    }
    return ((OsString) 0) ;
}

/* AUTODOC
 * (getpwnam USER)
 * Get users login informations.
 * This function querries the passwd file, and returns an object having the
 * follogin set of slots:
 *    name : USER login name (string), 
 *    uid  : USER user identifier (integer),
 *    gid  : USER group identifier (integer),
 *    dir  : USER home's directory (string),
 *    shell: USER login shell (string).
 * All those slots are read only.
 */

static OsObject OsSubrGetpwnam (uname)
    OsString uname ;
{
    struct passwd *entry ;
    OsShouldBeString (uname) ;
    if ( (entry = getpwnam (uname->string)) ) {
        struct passwd *found ;
        found = (struct passwd *) OsGCMallocAtomic (sizeof (struct passwd)) ;
        bcopy (entry, found, sizeof (struct passwd)) ;
        return ((OsObject) OsStructMake (pwd_sclass, OsPointerMake (found))) ;
    }
    return (OsA_False) ;
}

/* AUTODOC
 * (system COMMAND)
 * Launch COMMAND and wait for its completion.
 * The basic interface to launch commands from oscheme. It forks the provided
 * command and waits for its completion. It triggers the 'error:bad-syscall
 * if the underlying system call "system" fails.
 * It returns the exit status of the launched process.
 */

static OsObject OsSubrSystem (cmd)
    OsString cmd ;
{
    int status ;
    
    OsShouldBeString (cmd) ;
    OsEnvUpdate () ;
    if ( (status = system (cmd->string)) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("system")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) cmd, OsA_Nil)) ;
        /* NOT REACHED */
    }
    return ((OsObject) OsIntegerMake (status)) ;
}

/*
 * Working directory:
 */

/* AUTODOC
 * (chdir DIR)
 * CHange working directory to DIR.
 * This function change oscheme working directory to DIR. It triggers the
 * 'error:bad-syscall if the underlying system "chdir" call failed.
 */

static OsObject OsSubrChdir (newdir)
    OsString newdir ;
{
    OsShouldBeString (newdir) ;
    if ( chdir (newdir->string) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("chdir")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) newdir, OsA_Nil)) ;
    }
    return (OsA_True) ;
}

/* AUTODOC
 * (getcwd)
 * Get current working directoy.
 * This primitives returns oscheme current working directory, as a string. It
 * triggers the 'error:bad-syscall when the underlying system "getcwd" call
 * fails.
 */

static OsObject OsSubrGetcwd () {
    char tmp[MAXPATHLEN+1] ;

#ifdef OS_HAS_GETWD
    if ( ! getwd (tmp) )
#else
    if ( ! getcwd (tmp, MAXPATHLEN) )
#endif
    {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("getcwd")
                               , OsIntegerMake (error)
                               , OsA_Nil) ;
    }
    return ((OsObject) OsStringMake (tmp)) ;
}

/* AUTODOC
 * (machine-type)
 * Get the machine type.
 * This primitive returns the machine on which this version of oscheme was
 * compiled.
 */

static OsObject OsSubrMachineType () {
    static OsAtom osa_machine = (OsAtom) 0 ;
    if ( ! osa_machine )
        osa_machine = OsAtomMake (OsCurrentInterp()
                                  , OS_MACHINE_TYPE, OsA_Unbound) ;
    return ((OsObject) osa_machine) ;
}

/* AUTODOC
 * (system-type)
 * Get system type.
 * This pritmitive returns the system on which the version of oscheme was
 * compiled.
 */

static OsObject OsSubrSystemType () {
    static OsAtom osa_system = (OsAtom) 0 ;
    if ( ! osa_system )
        osa_system = OsAtomMake (OsCurrentInterp()
                                 , OS_SYSTEM_TYPE, OsA_Unbound) ;
    return ((OsObject) osa_system) ;
}

/* AUTODOC
 * (sleep SECONDS)
 * Wait until the given number of seconds has elapsed.
 * After having stoped for the given number of SECONDS, this function always
 * returns #t.
 */

static OsObject OsSubrSleep (secs)
    OsInteger secs ;
{
    OsShouldBeInteger (secs) ;
    sleep (secs->ival) ;
    return (OsA_True) ;
}

/* AUTODOC
 * (current-time)
 * Number of seconds since January the 1st, 1970.
 * See also the "date" module which provides utility functions to deal with
 * date.
 */

static OsObject OsSubrCurrentTime () {
    time_t lt ;
    time (&lt) ;
    return ((OsObject) OsIntegerMake (lt)) ;
}

/*
 * Initialize
 */

void OsUnixInitialize (OsInterp interp, int *pargc, char **argv) {
    os_env_initialize () ;
    return ;
}

void OsUnixSubrInitialize (OsInterp interp) {
    pwd_sclass_initialize () ;
    stat_sclass_initialize () ;
    OsSubrMake (interp, "exit", OsSubrExit, 1) ;
    OsSubrMake (interp, "putenv", OsSubrPutenv, 2) ;
    OsSubrMake (interp, "getenv", OsSubrGetenv, 1) ;
    OsSubrMake (interp, "file-exists?", OsSubrFileExistsP, 1) ;
    OsSubrMake (interp, "file-stat", OsSubrFileStat, 1) ;
    OsSubrMake (interp, "getpwnam", OsSubrGetpwnam, 1) ;
    OsSubrMake (interp, "getcwd", OsSubrGetcwd, 0) ;
    OsSubrMake (interp, "machine-type", OsSubrMachineType, 0) ;
    OsSubrMake (interp, "system-type", OsSubrSystemType, 0) ;
    OsSubrMake (interp, "sleep", OsSubrSleep, 1) ;
    OsSubrMake (interp, "current-time", OsSubrCurrentTime, 0) ;
    OsSecSubrMake (interp, "system", OsSubrSystem
                   , 1, OS_SECURITY_LEVEL_A) ;
    OsSecSubrMake (interp, "chdir", OsSubrChdir
                   , 1, OS_SECURITY_LEVEL_B) ;
    OsSecSubrMake (interp, "directory", OsSubrDirectory
                   , 1, OS_SECURITY_LEVEL_B) ;
    return ;
}
