;; edit-tree.scm
;; edit-tree.scm,v 1.1.1.1 1995/02/27 15:14:53 baird Exp
;; See the file COPYRIGHT for licensing conditions.
;; AUTODOC MODULE
;; A simple tree editor, using motif list widgets.

;; The osfb scripts provides a sample usage of this package.

;; We need to add these settings to fallback resources:

(xt:set-fallback-resources!
 (list->vector
  (append 
   (list
    "*lbox*l1.listSizePolicy: constant"
    "*lbox*l2.listSizePolicy: constant"
    "*lbox*l3.listSizePolicy: constant")
   (vector->list (xt:fallback-resources)))))

;; AUTODOC AUTOLOAD
;; (make-tree-editor NAME PARENT)
;; Create a tree editor named NAME inside PARENT widget.
;; This function creates a tree editor a la Next. This editor is made of three
;; lists widgets which are linked together, in order to display one full depth
;; of the tree.
;; Using this tree editor should be failrly esay. First, call this function
;; and get a tree-editor object. Then, refine it by implementing the following
;; set of messages:
;;    - (itemsAt: DEPTH) should return a vector containing all the items at
;;      given DEPTH in the tree. When implementing this message, you may use
;;      the tree-editor 'selection slot value.
;;    - (activate) the user want to "do" something with the current selection,
;;      this is the place to implement "do".
;;    - (selectionChanged) is invoked when the selection has changed. It 
;;      should return a string, to be displayed in the text entry on top of
;;      the lists.
;; Once you have implemented these messages, just invoke the (startup) method.
;; The tree-editor object defines the following user-callable stuff:
;;    - (selection) is a slot containing the current path in the tree, as a
;;      vector,
;;    - (maxDepth) is a slot limiting the maximum depth of the visible tree,
;;      which defaults to 10,
;;    - (startup) is a message implementing the tree editor initialization.
;; As an example, the following code implements a terse file browser with this
;; package:
;; 
;; (require "edit-tree")
;; 
;; (define (make-filebrowser factory root-path)
;;   (let* ((frame [factory frame: "frame" parent: [factory appToplevel]])
;;          (tree  (make-tree-editor "file-browser" frame)))
;;     ;; Refine the (itemsAt:) message
;;     (define-message tree (itemsAt: depth)
;;       (if (= depth 0)
;;           ;; Return the root path.
;;           (vector root-path)
;;           ;; If the given selection represents a directory, then list it to
;;           ;; get next level in the tree.
;;           (let ((path (apply path-append
;;                              (vector->list (subvector [self selection] 
;;                                                       0 (- depth 1))))))
;;             (and (file-exists? path)
;;                  (directory? path) 
;;                  (list->vector (directory path))))))
;;     ;; Launch the browser:
;;     [tree startup]
;;     tree) )
;;   
;; (define factory (motif:make-factory (getenv "DISPLAY") "foo" "Foo"))
;; 
;; (define fb (make-filebrowser factory (expand-path-for-tilde "~")))
;; [[factory appToplevel] realize]

(define make-tree-editor
  (let ((editor [obj:root clone]))
    [editor defineSlot: "list1"     init: #f]
    [editor defineSlot: "list2"     init: #f]
    [editor defineSlot: "list3"     init: #f]
    [editor defineSlot: "path"      init: #f]
    [editor defineSlot: "maxDepth"  init: 10]
    [editor defineSlot: "curDepth"  init: 0]
    [editor defineSlot: "selection" init: (vector)]

    ;; List widget utilities:
    ;; ----------------------
    ;; clear-list: clear the list widget content:
    (define (clear-list lst)
      [lst items: (vector) itemCount: 0])
    ;; move-items: move item from source widget list to dest
    (define (move-items source dest clear?)
      (let ((source-items [source items])
            (source-sel   [source selectedItems]))
        [dest items: source-items itemCount: (vector-length source-items)]
        ;; Preserve single selection, and its visibility:
        (and (> (vector-length source-sel) 0)
             (let ((item (vector-ref source-sel 0)))
               [dest selectItem: item callCallback: #f]
               [dest setItem: item]))
        ;; Clear if needed the source list
        (and clear? (clear-list source))) )

    ;; Internal messages:
    ;; ------------------
    ;; (new) build a new tree editor:
    (define-message editor (_new: name parent: parent)
      (let* ((new     [self clone])
             (factory [parent factory])
             (box     [factory rowColumn: name   parent: parent])
             (ctrl    [factory form:      "ctrl" parent: box])
             (lbox    [factory form:      "lbox" parent: box]))
        [box  orientation: "vertical"]
        ;; Make control buttons:
        (let ((left  [factory arrowButton: "left"  parent: ctrl])
              (path  [factory textField:  "path"  parent: ctrl])
              (right [factory arrowButton: "right" parent: ctrl]))
	  [new path: path]
	  ;; Geometry stuff:
	  [left arrowDirection: "ARROW_LEFT"
		topAttachment: "ATTACH_FORM" 
		bottomAttachment: "ATTACH_FORM"
		leftAttachment: "ATTACH_FORM"]
	  [path topAttachment: "ATTACH_FORM"
		bottomAttachment: "ATTACH_FORM"
		leftAttachment: "ATTACH_WIDGET" leftWidget: left
		rightAttachment: "ATTACH_WIDGET" rightWidget: right]
	  [right arrowDirection: "ARROW_RIGHT"
		 topAttachment: "ATTACH_FORM"
		 bottomAttachment: "ATTACH_FORM"
		 rightAttachment: "ATTACH_FORM"]
          [left defineCallback: "activate" to: (lambda (w) [new _shiftRight])]
          [right defineCallback: "activate" to: (lambda (w) [new _shiftLeft])])
        ;; Make the lists:
        (let ((l1   [factory scrolledList: "l1" parent: lbox])
              (l2   [factory scrolledList: "l2" parent: lbox])
              (l3   [factory scrolledList: "l3" parent: lbox]))
          [new list1: l1]
          [new list2: l2]
          [new list3: l3]
	  ;; Geometry stuff:
	  [lbox fractionBase: 3 width: 450 height: 200]
	  [[l1 parent] topAttachment:    "ATTACH_FORM"
		       bottomAttachment: "ATTACH_FORM"
		       leftAttachment:   "ATTACH_POSITION" leftPosition: 0
		       rightAttachment:  "ATTACH_POSITION" rightPosition: 1]
	  [[l2 parent] topAttachment:    "ATTACH_FORM"
		       bottomAttachment: "ATTACH_FORM"
		       leftAttachment:   "ATTACH_POSITION" leftPosition: 1
		       rightAttachment:  "ATTACH_POSITION" rightPosition: 2]
	  [[l3 parent] topAttachment:    "ATTACH_FORM"
		       bottomAttachment: "ATTACH_FORM"
		       leftAttachment:   "ATTACH_POSITION" leftPosition: 2
		       rightAttachment:  "ATTACH_POSITION" rightPosition: 3]
          ;; Hack the three lists in the same way:
          (define (make-select-cb level)
            (lambda (w) 
              (let ((selected [w selectedItems]))
                (when (> (vector-length selected) 0)
                      [new _select:  (vector-ref [w selectedItems] 0)
                           atLevel: level]))))
          (define (make-activate-cb) (lambda (w) [new activate]))
          [l1 scrollBarDisplayPolicy: "static"]
          [l1 selectionPolicy: "SINGLE_SELECT"]
          [l1 defineCallback: "singleSelection" to: (make-select-cb 0)]
          [l1 defineCallback: "defaultAction" to: (make-activate-cb)]
          [l2 scrollBarDisplayPolicy: "static"]
          [l2 selectionPolicy: "SINGLE_SELECT"]
          [l2 defineCallback: "singleSelection" to: (make-select-cb 1)]
          [l2 defineCallback: "defaultAction" to: (make-activate-cb)]
          [l3 scrollBarDisplayPolicy: "static"]
          [l3 selectionPolicy: "SINGLE_SELECT"]
          [l3 defineCallback: "singleSelection" to: (make-select-cb 2)]
          [l3 defineCallback: "defaultAction" to: (make-activate-cb)]
          new)))
    ;; (_shiftLeft) shoft lists content to the left
    (define-message editor (_shiftLeft)
      (when (and (< [self curDepth] [self maxDepth])
                 (> (vector-length [self selection]) (+ [self curDepth] 2)))
            ;; If list3 cannot be filled, shifLeft has no effect:
            (let* ((d (+ [self curDepth] 1 2))
                   (s [self selection])
                   (f [self itemsAt: d]))
              (when f
                    [self curDepth: (+ 1 [self curDepth])]
                    ;; shift lists l2->l1 and l3->l2
                    (move-items [self list2] [self list1] #f)
                    (move-items [self list3] [self list2] #f)
                    ;; fill list3
                    [[self list3] items: f itemCount: (vector-length f)]
                    (and (> (vector-length s) d)
                         [[self list3] selectItem: (vector-ref s d)
                                       callCallback: #f])))))
    ;; (_shiftRight) shift list content to the right
    (define-message editor (_shiftRight)
      (when (> [self curDepth] 0)
            ;; shift lists l2->l3, l1->l2
            (move-items [self list2] [self list3] #f)
            (move-items [self list1] [self list2] #f)
            [self curDepth: (- [self curDepth] 1)]
            ;; fill list1, fill should always be available here
            (let ((fill [self itemsAt: [self curDepth]]))
              [[self list1] items: fill itemCount: (vector-length fill)]
              [[self list1] selectItem: (vector-ref [self selection] 
                                                    [self curDepth])
                            callCallback: #f]
	      [[self list1] setItem: (vector-ref [self selection]
						 [self curDepth])])))
    ;; (_select:atLevel) when some item is selected in a list
    ;; This is the main stuff: when an item is selected in list at LEVEL, then
    ;; try to extend the path "on the right"
    (define-message editor (_select: item atLevel: level)
      ;; Update our view of the current selected path:
      (let ((l (+ [self curDepth] level))
            (s [self selection]))
        (if (= l 0) 
            [self selection: (vector item)]
            [self selection: (vector-append! (subvector s 0 (- l 1)) item)]))
      ;; Update GUI:
      (cond ((= level 0)
             ;; clear right-hand list
             (clear-list [self list3])
             ;; re-fill middle list
             (let ((items [self itemsAt: (+ [self curDepth] 1)]))
               (and items
                    [[self list2] items: items 
                                  itemCount: (vector-length items)]
                    [[self list2] selectedItems: (vector)
                                  selectedItemCount: 0])))
            ((= level 1)
             ;; re-fill right-hand list:
             (let ((items [self itemsAt: (+ [self curDepth] 2)]))
               (and items
                    [[self list3] items: items
                                  itemCount: (vector-length items)]
                    [[self list3] selectedItems: (vector)
                                  selectedItemCount: 0])))
            ((= level 2)
             ;; just shiftLeft auto-magically !
             [self _shiftLeft]))
      [self _selectionChanged])
    ;; (_selectionChanged) the tree editor has a new selection
    (define-message editor (_selectionChanged)
      (let ((value [self selectionChanged]))
        (and value [[self path] value: value])))

    ;; User callable messages:
    ;; -----------------------
    ;; (startup) Run the tree, starting at fiorst column
    (define-message editor (startup)
      (let ((initial-items [self itemsAt: 0]))
        [[self list1] items: initial-items 
                      itemCount: (vector-length initial-items)]))
    ;; User redefinable messages:
    ;; --------------------------
    ;; (selectionChanged) redisplay the text value to match current selection
    (define-message editor (selectionChanged)
      (apply path-append (vector->list [self selection])))
    ;; (itemsAt) get back the list of items for absolute depth AT
    (define-message editor (itemsAt: depth)
      (error "make-tree-editor: \"itemsAt:\" not refined for %0\n" self))
    ;; (activate) current selection has been activated
    (define-message editor (activate)
      (format (current-output-port)
              "editor[activate]: %0\n" [self selection]))

    (lambda (name parent)
      (let ((new [editor _new: name parent: parent])) new))) )

(provide "edit-tree")

