/***************************************************************************
 * main.c
 * main.c,v 1.4 1995/03/17 14:27:22 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * C level interface to the Motif toolkit.
 * This module provides the raw interface to Motif, using oscheme prototype
 * based object model.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

#include <X11/Intrinsic.h>
#include <X11/Xmu/Editres.h>

#include <Xm/Xm.h>

#include "os_motif.h"

static XtAppContext xtapp ;

/*
 * This one is expanded so we have a variable (debugger) on every step
 * os the expansion/compilation/evaluation phases
 */

static void stdin_handler ()
{
    OsObject obj, error ;
    obj = OsCProtectExecute (OsCurrentInterp()
                             , (OsObject (*)()) OsRead
                             , (OsObject) OsCurrentInputPort
                             , &error) ;
    if ( error ) {
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
    } else {
        OsExprEval (obj, 0, (OsObject *) 0) ;
    }
    OsOutputPortWriteCString (OsCurrentOutputPort, "=> ") ;
    OsOutputPortFlush (OsCurrentOutputPort) ;
    return ;
}

/* AUTODOC
 * (motif:make-factory DISPLAY NAME CLASS)
 * Create a widget factory for DISPLAY, named NAME, of (Xt)class CLASS.
 * This primitive opens the given DISPLAY, and create a factory object whose
 * application name will be NAME, and application class CLASS (both NAME and
 * CLASS have their Xt meaning here).
 * This factory has one method per motif widget class, which, when invoked
 * will return an instance (widget) of the given class.
 * Additional methods are:
 *    (appName): to retreive the application name given at creation time,
 *    (appClass): to retreieve the class,
 *    (appDisplay): to retreive the display,
 *    (appToplevel): to get the underlying top level widget.
 * 
 * For example, to create the famous "hello world" example:
 * ;; Build a widget hierarchy, ending up with a pushButton widget:
 * (define factory (motif:make-factory (getenv "DISPLAY") "hello" "Hello"))
 * (define frame  [factory frame: "frame" parent: [factory appToplevel]])
 * (define button [factory pushButton: "Say Hello !" parent: frame])
 * ;; Make the button say hello:
 * [button defineCallback: "activate"
 *         to: (lambda (widget) (display "Hello World\n"))]
 * ;; Realize widgets:
 * [[factory appToplevel] realize]
 * 
 * As the above example shows, widgets are full objects. They always provide
 * the following set of messages:
 *    (realize): to realize (in the Xt sense) the given widget,
 *    (defineCallback:to:) to define the callback whose name is given to the
 *        given function. When the callback is triggered, it is always provided
 *        with the widget object.
 *    (widgetClass): returns a string describing the widget class.
 *    (managed): is a slot. When set to #t the widget is managed, otherwise, it
 *        unmanaged. When created, widgets are always managed.
 *    (callAction:with): call the action given by its name (as a scheme symbol)
 *        and the set of parameters packed in the WITH vector:
 *            [text callAction: 'delete-to-end-of-line with: (vector)]
 * More over, the widget object gets their (forward) method implementation
 * refined in the following manner: each time a widget doesn't understand a
 * message, it looks up the message name, and if it matches a resource name,
 * then set (or get) the provided resource. Hence to set some button background
 * one will use:
 *     [button background: "red"]
 * To get the button width:
 *     [button width]
 * Multiple resources can be set at once, for example:
 *     [widget topAttachment: "ATTACH_WIDGET" topWidget: <widget>]
 *
 * Some widgets get special methods attached to them, at creation time:
 * List widgets (scrolledList, list):
 *    (addItem:at:) Add ITEM at given POSITION in list,
 *    (selectPosition:callCallback:) Select item at given position. If
 *        callCallback isn't #f, then trigger the selection callback,
 *    (selectItem:callCallback) Same as above, ecxept that item is provided by
 *        its string value instead of its position in the list,
 *    (setItem): Make the given ITEM be in visible.
 *
 * Text widgets (scrolledText, text and textField):
 *    (insertionPosition) get the current insertion position,
 *    (insertionPosition:) set the current insertion position,
 *    (lastPosition): get the last position in the text,
 *    (insert:at:) Insert given STRING at given POSITION,
 *    (replaceFrom:to:with:) Replace the text from FROM to TO with WITH,
 *    (enableDisplay:) Enable/disable text redisplay.
 * Finally, motif StringTables are converted as vector of oscheme strings, and
 * numerical (i.e: int, short, Cardinal, Position, Dimension...) resources
 * are converted to oscheme integers.
 */

static OsObject OsSubrMotifMakeFactory (os_dpy, os_name, os_class)
    OsString os_dpy, os_name, os_class ;
{
    extern OsObject MotifFactoryMake () ;
    OsShouldBeString (os_dpy) ;
    OsShouldBeString (os_name) ;
    OsShouldBeString (os_class) ;
    return (MotifFactoryMake (xtapp, os_dpy, os_name, os_class)) ;
}

static String *fallback_resources = (String *) 0 ;

/* AUTODOC
 * (xt:fallabck-resources)
 * Retreieve the current set of Xt fallback resources.
 * This function returns the current set of fallback resources, as a vector of
 * lines, following the X resource file syntax.
 */

static OsObject OsSubrXtFallbackResources ()
{
    OsVector resources ;
    String   *p = fallback_resources ;
    
    if ( ! fallback_resources )
        return ((OsObject) OsVectorMake(0)) ;
    resources = OsVectorMake (0) ;
    while ( *p ) {
        (void) OsVectorFastAppend (resources, (OsObject) OsStringMake (*p)) ;
        p++ ;
    }
    return ((OsObject) resources) ;
}

/* AUTODOC
 * (xt:set-fallback-resources! RESOURCES)
 * Install the vector of resources RESOURCES as fallback resources.
 * This function install the given set of RESOURCES, provided as a vector of
 * string following the X resource syntax, as the Xt fallback resources.
 */

static OsObject OsSubrXtSetFallbackResources (resources)
    OsVector resources ;
{
    unsigned int i ;
    OsShouldBeVector (resources) ;
    fallback_resources = (String *) OsGCMalloc (sizeof (String) 
                                                * resources->csize + 1) ;
    for (i = 0 ; i < resources->csize ; i++) {
        OsString rline = (OsString) resources->cells[i] ;
        OsShouldBeString (rline) ;
        fallback_resources[i] = (String) rline->string ;
    }
    fallback_resources[i] = (String) 0 ;
    XtAppSetFallbackResources (xtapp, fallback_resources) ;
    return (OsA_True) ;
}

/* Timeout management:
 * Needs protection against thunk (because it can be reached only through Xt
 * memory which is not allocated using Boehm's GC.... */
 
static OsVector timeout_protect = (OsVector) 0 ;
static OsVector input_protect   = (OsVector) 0 ;

static void xt_timeout_handler (tclos_ptr, ptimer)
    XtPointer    tclos_ptr ;
    XtIntervalId *ptimer ;
{
    OsVector tclos = (OsVector) tclos_ptr ;
    OsObject error = (OsObject) 0 ;
    OsObject again ;

    again = OsCodeBlockApplyWithArray ((OsInterp) tclos->cells[0]
                                       , (OsCodeBlock) tclos->cells[1]
                                       , 0, (OsObject *) 0
                                       , &error) ;
    if ( error ) {
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
    } else if ( again != OsA_False ) {
        XtAppAddTimeOut (xtapp
                         , ((OsInteger) tclos->cells[2])->ival
                         , xt_timeout_handler, tclos_ptr) ;
    } else {
        OsVectorFastRemoveq (timeout_protect, (OsObject) tclos) ;
    }
    return ;
}

static void xt_input_handler (XtPointer tclos_ptr, int *pfd, XtInputId *pid) {
    OsVector tclos = (OsVector) tclos_ptr ;
    OsObject error = (OsObject) 0 ;
    OsObject again ;

    again = OsCodeBlockApplyWithArray ((OsInterp) tclos->cells[0]
                                       , (OsCodeBlock) tclos->cells[1]
                                       , 1, &tclos->cells[2]
                                       , &error) ;
    if ( error ) {
        OsSendDisplay (OsErrorFormat ((OsVector) error), OsCurrentOutputPort) ;
        OsOutputPortWriteChar (OsCurrentOutputPort, '\n') ;
        OsOutputPortFlush (OsCurrentOutputPort) ;
    }
    if ( error || (again == OsA_False) ) {
        XtRemoveInput (*pid) ;
        OsVectorFastRemoveq (input_protect, (OsObject) tclos) ;
    }
    return ;
}

/* AUTODOC
 * (xt:add-timeout TIME THUNK)
 * Install THUNK to be executed after TIME milliseconds.
 * This primitive will make THUNK execute after TIME milliseconds have elapsed.
 * If the provided THUNK returns #t when executed, then it is automatically
 * reinstalled (after being executed) with the same delay.
 * The THUNK gets invoked within the interpreter context it has been setted.
 */

static OsObject OsSubrXtAddTimeout (OsInteger time, OsCodeBlock thunk) {
    OsVector protection ;
    
    OsShouldBeInteger (time) ;
    OsMustBeBlock (thunk) ;
    protection = OsVector3Make ((OsObject) OsCurrentInterp()
                                , (OsObject) thunk
                                , (OsObject) time) ;
    if ( ! timeout_protect )
        timeout_protect = OsVectorMake (0) ;
    (void) OsVectorFastAppend (timeout_protect, (OsObject) protection) ;
    XtAppAddTimeOut (xtapp
                     , time->ival
                     , xt_timeout_handler, (XtPointer) protection) ;
    return (OsA_True) ;
}

static OsObject OsSubrXtAddInput (OsObject io_port, OsCodeBlock thunk) {
    int      fd ;
    XtPointer condition ;
    OsVector protection ;

    OsMustBeClosedBlock (thunk) ;
    if ( OsIsInputPort (io_port) ) {
        fd        = ((OsInputPort) io_port)->medium.fd ;
        condition = (XtPointer) XtInputReadMask ;
    } else if ( OsIsOutputPort (io_port) ) {
        fd        = ((OsOutputPort) io_port)->medium.fd ;
        condition = (XtPointer) XtInputWriteMask ;
    } else {
        (void) OsErrorTrigger (1, OsE_BadType
                               , io_port
                               , OsList2Make ((OsObject) OsInputPortClass
                                              , (OsObject) OsOutputPortClass));
    }
    protection = OsVector3Make ((OsObject) OsCurrentInterp()
                                , (OsObject) thunk
                                , io_port) ;
    if ( ! input_protect )
        input_protect = OsVectorMake (0) ;
    (void) OsVectorFastAppend (input_protect, (OsObject) protection) ;
    XtAppAddInput (xtapp
                   , fd
                   , condition
                   , xt_input_handler
                   , (XtPointer) protection) ;
    return (OsA_True) ;
}

/* AUTODOC
 * (flush-widget WIDGET)
 * Flush the connection to WIDGET's X display.
 * This function allows to update GUI while executing Scheme code. It will
 * apply XtFlush to the display associated to the given WIDGET's display.
 */

static OsObject OsSubrXtFlushWidget (OsEObj os_widget) {
    Widget widget ;

    OsShouldBeEObj (os_widget) ;
    widget = OsCvtOsWidgetToWidget (os_widget) ;
    XSync (XtDisplay(widget), False) ;
    return (OsA_True) ;
}

static void before_profile_hook (OsInterp interp) {
    Os_osmotif_initialize (interp) ;
    OsSubrMake (interp, "motif:make-factory", OsSubrMotifMakeFactory, 3) ;
    OsSubrMake (interp, "xt:fallback-resources", OsSubrXtFallbackResources, 0);
    OsSubrMake (interp, "xt:add-timeout", OsSubrXtAddTimeout, 2) ;
    OsSubrMake (interp, "xt:add-input", OsSubrXtAddInput, 2) ;
    OsSubrMake (interp
                , "xt:set-fallback-resources!"
                , OsSubrXtSetFallbackResources, 1);
    OsSubrMake (interp, "xt:flush-widget", OsSubrXtFlushWidget, 1) ;
    XtToolkitInitialize () ;
    xtapp = XtCreateApplicationContext () ;
    XtAppAddInput (xtapp
                   , fileno (stdin)
                   , (XtPointer) XtInputReadMask
                   , stdin_handler
                   , (XtPointer) 0) ;
    Xt_converters_initialize () ;
    OsPathProvides (interp, OsStringMake ("motif")) ;
    return ;
}

int main (int argc, char **argv) {
    OsInterp        interp ;
    
    interp = (OsInterp) OsInitialize (&argc, argv, before_profile_hook) ;
    OsOutputPortWriteCString (OsCurrentOutputPort, "=> ") ;
    OsOutputPortFlush (OsCurrentOutputPort) ;
    XtAppMainLoop (xtapp) ;
}
