/***************************************************************************
 * process.c
 * proc.c,v 1.2 1995/03/12 02:59:04 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * AUTODOC MODULE
 * UNIX process management with scheme. This module provides a scheme
 * interface to UNIX process.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <sys/time.h>
#include <sys/types.h>
#ifdef OS_HAS_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#include <signal.h>
#ifdef OS_HAS_RLIMIT
#include <sys/time.h>
#include <sys/resource.h>
#endif
#ifdef OS_HAS_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef OS_HAS_SYS_SELECT_H
#include <sys/select.h>
#endif

extern int errno ;

#include <oscheme.h>

static OsHashtable os_processes   = (OsHashtable) 0 ;
static OsEObj      process_sclass = (OsEObj) 0 ;

static OsEObjSel   process_sel_rpid     = (OsEObjSel) 0 ;
static OsEObjSel   process_sel_wstatus  = (OsEObjSel) 0 ;
static OsEObjSel   process_sel_rcommand = (OsEObjSel) 0 ;
static OsEObjSel   process_sel_wcommand  = (OsEObjSel) 0 ;

/*
 * Internally used functions:
 */

static void child_death_handler (signo)
    int signo ;
{
    int    status ;
    int    pid ;
    OsEObj proc ;
    
#ifndef OS_HAS_WAIT3
    if ( (! (pid = wait (&status))) || ( ! status) ) {
#else 
    if ( (! (pid = wait3 (&status, WNOHANG, 0))) || ( ! status) ) {
#endif /* OS_HAS_WAIT3 */
        return ;
    }
    if ( ! (proc = (OsEObj) OsHashtableGet (os_processes, (void *) pid)) )
        return ;
    OsEObjSendMsg1 (proc, proc
                    , process_sel_wstatus
                    , (OsObject) OsIntegerMake(status));
    return ;
}

static void process_finalizer (proc)
    OsEObj proc ;
{
    int pid = (int) OsHashtableFindKey (os_processes, proc) ;
    if ( pid )
        OsHashtableRemove (os_processes, (void *) pid) ;
    return ;
}

/*
 * Raw message implementations:
 */

static OsObject process_wait (closure, self, sel)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
{
    OsInteger pid  = ((OsInteger) OsEObjSendMsg0 (self, self
                                                  , process_sel_rpid)) ;
    int       status ;
    OsInteger os_status ;
    
    if ( waitpid (pid->ival, &status, 0) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("waitpid")
                               , OsIntegerMake (error)
                               , OsConsMake ((OsObject) pid, OsA_Nil)) ;
    }
    os_status = OsIntegerMake (status) ;
    OsEObjSendMsg1 (self, self, process_sel_wstatus, (OsObject) os_status) ;
    return ((OsObject) os_status) ;
}

static OsObject process_kill (closure, self, sel, signo)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsInteger signo ;
{
    OsInteger pid  = (OsInteger) OsEObjSendMsg0 (self, self
                                                 , process_sel_rpid) ;
    OsShouldBeInteger (signo) ;

    if ( kill (pid->ival, signo->ival) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("kill")
                               , OsIntegerMake (error)
                               , OsList2Make ((OsObject) pid
                                              , (OsObject) signo)) ;
    }
    return ((OsObject) signo) ;
}

static OsObject process_display_to (closure, self, sel, port)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
    OsOutputPort port ;
{
    OsObject object ;
    
    OsOutputPortWriteCString (port, "#[Process (") ;
    object = OsEObjSendMsg0 (self, self, process_sel_rpid) ;
    OsSendDisplay (object, port) ;
    OsOutputPortWriteChar (port, ')') ;
    OsOutputPortWriteChar (port, ' ') ;
    object = OsEObjSendMsg0 (self, self, process_sel_rcommand) ;
    OsSendDisplay (object, port) ;
    OsOutputPortWriteChar (port, ']') ;
    return ((OsObject) self) ;
}

static void process_initialize () {
    os_processes   = OsWeakHashtableMake (0
                                          , (OsHashFct) OsHashInteger
                                          , (OsCmpFct) OsCmpInteger) ;
    process_sclass = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                              , (OsEObj) OsEObjRoot
                                              , OsEObjSelClone) ;

/*
 * process slots (default to nil here):
 */
    process_sel_rcommand = OsEObjSelDefine ("command") ;
    process_sel_wcommand = OsEObjSelDefine ("command:") ;
    OsEObjAllocateSlot (process_sclass
                        , process_sel_rcommand
                        , process_sel_wcommand
                        , OsA_Unbound) ;
    process_sel_rpid    = OsEObjSelDefine ("pid") ;
    OsEObjAllocateSlot (process_sclass
                        , process_sel_rpid
                        , (OsEObjSel) 0
                        , (OsObject) OsIntegerMake (0)) ;
    process_sel_wstatus = OsEObjSelDefine ("status:") ;
    OsEObjAllocateSlot (process_sclass
                        , OsEObjSelDefine ("status")
                        , process_sel_wstatus
                        , (OsObject) OsIntegerMake (0)) ;
    OsEObjAllocateSlot (process_sclass
                        , OsEObjSelDefine ("input")
                        , (OsEObjSel) 0
                        , OsA_Unbound) ;
    OsEObjAllocateSlot (process_sclass
                        , OsEObjSelDefine ("output")
                        , (OsEObjSel) 0
                        , OsA_Unbound) ;
    OsEObjAllocateSlot (process_sclass
                        , OsEObjSelDefine ("error")
                        , (OsEObjSel) 0
                        , OsA_Unbound) ;
/*
 * process predefined messages:
 */
    OsEObjDefineMsg (process_sclass
                     , OsEObjSelDefine ("display:")
                     , process_display_to
                     , (void *) 0) ;
    OsEObjDefineMsg (process_sclass
                     , OsEObjSelDefine ("wait")
                     , process_wait
                     , (void *) 0) ;
    OsEObjDefineMsg (process_sclass
                     , OsEObjSelDefine ("kill:")
                     , process_kill
                     , (void *) 0) ;
    return ;
}

/* AUTODOC (AUTOLOAD "process")
 * (make-process COMMAND ARGS*)
 * Launch a process executing COMMAND with given ARGS*.
 * This process interface is available only on UNIX system. It uses the object
 * model: the process interface creates an object with the following set of
 * slots:
 *    pid    : contains the process pid (read only),
 *    status : contains the process status (read/write)
 *    input  : contains the process input stream (read only)
 *    output : contains the process output stream (read only)
 *    error  : contains the process error stream (read only)
 * For each process object, the following set of messages is implemented:
 *    wait   : wait for the process to terminate
 *    kill:  : sends process signal SIGNO
 * As a brief example, consider:
 * > (define proc (make-process "cat"))
 * ;; #[Process Object]
 * > (display "Hello World.\n" [proc input])
 * ;; #[output port]
 * > (select -1 [proc input])
 * ;; #(...)
 * > (read-line [proc output])
 * ;; "Hello World"
 * > [proc kill: 9]
 * ;; 9
 * > [proc status]
 * ;; 9
 */
 
static OsObject OsSubrMakeProcess (unsigned int argc, OsString *argv) {
    int    pid, fd_in[2], fd_out[2], fd_err[2] ;
    OsEObj proc ;
    int    n ;
    char   **cargv ;
    
    /* Check args, built arg array */
    for ( n = 0 ; n < argc ; n++ ) {
        OsShouldBeString (argv[n]) ;
    }
    cargv = (char **) OsGCMallocAtomic ((argc+1) * sizeof (char *)) ;
    for ( n = 0 ; n < argc ; n++ ) {
        cargv[n] = argv[n]->string ;
    }
    cargv[n] = NULL ;
    OsEnvUpdate () ;
    /* Prepare pipes, fork */
    pipe (fd_in) ; pipe (fd_out) ; pipe (fd_err) ;
#ifdef OS_HAS_FORK
    if ( pid = fork () )
#else
    if ( pid = vfork () )
#endif
    { 	/* father */
        FILE *fin = fdopen (fd_in[1], "w") ;
        close (fd_in[0]) ;
        close (fd_out[1]) ;
        close (fd_err[1]) ;
        proc = (OsEObj) OsEObjSendMsg0 (process_sclass
                                        , process_sclass
                                        , OsEObjSelClone) ;
        /* Initialize slots:
         * By initializing them in the same order they were created, we share
         * our parent class, this makes a big difference... */
        OsEObjAllocateSlot (proc
                            , OsEObjSelDefine ("pid") 
                            , (OsEObjSel) 0
                            , (OsObject) OsIntegerMake (pid)) ;
        OsEObjSendMsg1 (proc, proc
                        , process_sel_wcommand
                        , (OsObject) argv[0]) ;
        OsEObjSendMsg1 (proc, proc
                        , process_sel_wstatus
                        , (OsObject) OsIntegerMake (0));
        OsEObjAllocateSlot (proc
                            , OsEObjSelDefine ("input")
                            , (OsEObjSel) 0
                            , (OsObject) OsOutputPortMakeFromFile (fin)) ;
        OsEObjAllocateSlot (proc
                            , OsEObjSelDefine ("output")
                            , (OsEObjSel) 0
                            , (OsObject) OsInputPortMakeFromFileno(fd_out[0]));
        OsEObjAllocateSlot (proc
                            , OsEObjSelDefine ("error")
                            , (OsEObjSel) 0
                            , (OsObject) OsInputPortMakeFromFileno(fd_err[0]));

        (void) OsHashtablePut (os_processes, (void *) pid, proc) ;
        OsGCRegisterFinalizer (proc, process_finalizer) ;
        return ((OsObject) proc) ;
    } else if ( ! pid ) {	/* child */
        int fd ;
        dup2 (fd_in[0], 0) ;
        dup2 (fd_out[1], 1) ;
        dup2 (fd_err[1], 2) ;
#ifdef OS_HAS_RLIMIT
        {
            struct rlimit rlp ;
            getrlimit (RLIMIT_NOFILE, &rlp) ;
            for (fd = rlp.rlim_cur ; --fd > 2 ; close (fd))
                ;
        }
#else
        for (fd = getdtablesize() ; --fd > 2 ; close (fd))
	    ;
#endif
	execvp (cargv[0], cargv) ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("execvp")
                               , OsIntegerMake (errno)
                               , OsConsMake ((OsObject) argv[0], OsA_Nil)) ;
    } else {			/* error */
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("fork")
                               , OsIntegerMake (error)
                               , OsA_Nil) ;
    }
    /* NOT REACHED */
}

/* AUTODOC (AUTOLOAD "process")
 * (select TIMEOUT PORTS*)
 * Wait for ports to be ready for IO, or TIMEOUT to expire.
 * Given any kind of streams, this functions blocks until either:
 *    - some input port of PORTS is (are) ready to be read,
 *    - some output port of PORTS is (are) ready to be written,
 *    - the TIMEOUT (in micro-secs) expire.
 * It than returns a vector containing two lists: one for ready input ports,
 * and the other for ready output ports, both being null if the TIMEOUT has
 * expired.
 * Finally, if the TIMEOUT is negative, than the primitive blocks for ever
 * until some port is ready, or if it is 0, the primitive just pool the set of
 * input ports.
 */

static OsObject OsSubrSelect (unsigned int argc, OsObject *argv) {
    struct timeval timeout, *ptimeout = &timeout ;
    int            mxfd = 0, fd = 0, nfd = 0, i = 0 ;
    fd_set         rfds, wfds ;
    OsCons         os_rfds = (OsCons) OsA_Nil, *p_os_rfds = &os_rfds ;
    OsCons         os_wfds = (OsCons) OsA_Nil, *p_os_wfds = &os_wfds ;
    
    /* Get timeout */
    if ( argc < 1 )  
        (void) OsErrorTrigger (1, OsE_ArityShouldBe
                               , OsIntegerMake (1)
                               , OsIntegerMake (0)) ;
    bzero (ptimeout, sizeof (struct timeval)) ;
    OsMustBeInteger (argv[0]) ;
    if ( ((OsInteger) argv[0])->ival < 0 ) {
        ptimeout = (struct timeval *) 0 ;
    } else {
        timeout.tv_usec = ((OsInteger) argv[0])->ival ;
    }
    argc-- ; argv++ ;
    /* Get rfds, wfds */
    FD_ZERO (&rfds) ;
    FD_ZERO (&wfds) ;
    for (i = 0 ; i < argc ; i++ ) {
        if ( OsIsInputPort (argv[i]) ) {
            OsInputPort port = (OsInputPort) argv[i] ;
            if ( (fd = port->os_ready (port)) < 0 ) {
                *p_os_rfds = OsConsMake ((OsObject) port, OsA_Nil) ;
                p_os_rfds  = (OsCons *) &((*p_os_rfds)->cdr) ;
                argv[i]    = (OsObject) 0 ;
            } else {
                FD_SET (fd, &rfds) ;
            }
        } else if ( OsIsOutputPort (argv[i]) ) {
            OsOutputPort port = (OsOutputPort) argv[i] ;
            if ( (fd = port->os_ready (port)) < 0 ) {
                *p_os_wfds = OsConsMake ((OsObject) port, OsA_Nil) ;
                p_os_wfds  = (OsCons *) &((*p_os_wfds)->cdr) ;
            } else {
                FD_SET (fd, &wfds) ;
            }
        } else {
            (void)OsErrorTrigger(1, OsE_BadType
                                 , argv[i]
                                 , OsList2Make ((OsObject)OsOutputPortClass
                                                , (OsObject)OsInputPortClass));
        }
        if ( fd >= mxfd )
            mxfd = fd + 1 ;
    }
    /* select, finally */
    if ( (nfd = select (mxfd, &rfds, &wfds, (fd_set *) 0, ptimeout)) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("select")
                               , OsIntegerMake (error)
                               , OsList2Make ((OsObject) os_rfds
                                              , (OsObject) os_wfds)) ;
    }
    for (i = 0 ; nfd ; i++) {
        if ( argv[i] ) {	/* see above hack to get read of readies */
            if ( OsIsInputPort (argv[i]) ) {
                OsInputPort port = (OsInputPort) argv[i] ;
                if ( FD_ISSET (port->medium.fd, &rfds) ) {
                    nfd-- ;
                    *p_os_rfds = OsConsMake ((OsObject) port, OsA_Nil) ;
                    p_os_rfds  = (OsCons *) &((*p_os_rfds)->cdr) ;
                }
            } else if ( OsIsOutputPort (argv[i]) ) {
                OsOutputPort port = (OsOutputPort) argv[i] ;
                if ( FD_ISSET (fileno (port->medium.fp), &wfds) ) {
                    nfd-- ;
                    *p_os_wfds = OsConsMake ((OsObject) port, OsA_Nil) ;
                    p_os_wfds  = (OsCons *) &((*p_os_wfds)->cdr) ;
                }
            }
        }
    }
    return ((OsObject) OsVector2Make ((OsObject) os_rfds, (OsObject) os_wfds));
}

/* AUTODOC (AUTOLOAD "process")
 * (kill-process PID SIGNO)
 * Sends process PID signal SIGNO.
 * This primitive sends signal SIGNO to the given process. It triggers the
 * error:bad-syscall error if the underlying killsystem call fails.
 */

static OsObject OsSubrKillProcess (pid, signo)
    OsInteger pid, signo ;
{
    pid   = OsShouldBeInteger (pid) ;
    signo = OsShouldBeInteger (signo) ;
    if ( kill (pid->ival, signo->ival) < 0 ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("kill")
                               , OsIntegerMake (error)
                               , OsList2Make ((OsObject) pid
                                              , (OsObject) signo)) ;
    } else {
        return (OsA_True) ;
    }
    /* NOT REACHED */
}

/* AUTODOC (AUTOLOAD "process")
 * (with-input-from-process COMMAND THUNK)
 * Launch COMMAND, and invoke THUNK with current-input to process output.
 */

static OsObject OsSubrWithInputFromProcess (command, thunk)
    OsString    command ;
    OsCodeBlock thunk ;
{
    FILE        *finput ;
    OsInputPort iport, oldport ;
    OsObject    value, error ;
    
    OsShouldBeString (command) ;
    OsMustBeBlock (thunk) ;
    if ( ! (finput = popen (command->string, "r")) ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("popen")
                               , OsIntegerMake (error) 
                               , OsConsMake ((OsObject) command, OsA_Nil)) ;
        /* NOT REACHED */
    }
    iport = OsInputPortMakeFromFileno (fileno (finput)) ;
    oldport = OsCurrentInputPort ;
    OsCurrentInputPort = iport ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , thunk
                                       , 0, (OsObject *) 0
                                       , &error) ;
    /* We both pclose, and InputPortClose...this might be ok, if no file
     * operations occur in between (?) */
    pclose (finput) ; 
    OsCurrentInputPort = oldport ;
    OsInputPortClose (iport) ;
    if ( error ) 
        os_uncatched_error (error) ;
    else
        return (value) ;
}

/* AUTODOC (AUTOLOAD "process")
 * (with-output-to-process COMMAND THUNK)
 * Launches COMMAND, and invoke THUNK with current-output to process input.
 */

static OsObject OsSubrWithOutputToProcess (command, thunk)
    OsString    command ;
    OsCodeBlock thunk ;

{
    FILE        *foutput ;
    OsOutputPort oport, oldport ;
    OsObject    value, error ;
    
    OsShouldBeString (command) ;
    OsMustBeBlock (thunk) ;
    if ( ! (foutput = popen (command->string, "w")) ) {
        int error = errno ;
        (void) OsErrorTrigger (1, OsE_BadSyscall
                               , OsStringMake ("popen")
                               , OsIntegerMake (error) 
                               , OsConsMake ((OsObject) command, OsA_Nil)) ;
        /* NOT REACHED */
    }
    oport = OsOutputPortMakeFromFile (foutput) ;
    oldport = OsCurrentOutputPort ;
    OsCurrentOutputPort = oport ;
    value = OsCodeBlockApplyWithArray (OsCurrentInterp()
                                       , thunk
                                       , 0, (OsObject *) 0
                                       , &error) ;
    /* We both pclose, and OutputPortClose...this might be ok, if no file
     * operations occur in between (?) */
    pclose (foutput) ; 
    OsCurrentOutputPort = oldport ;
    OsOutputPortClose (oport) ;
    if ( error ) 
        os_uncatched_error (error) ;
    else
        return (value) ;
}

void Os_process_initialize (OsInterp interp) {
    process_initialize () ;
    OsCSignalHandle (SIGCHLD, child_death_handler) ;
    OsSubrMake (interp,"select", OsSubrSelect, OS_NARY) ;
    OsSecSubrMake (interp
                   , "make-process", OsSubrMakeProcess
                   , OS_NARY, OS_SECURITY_LEVEL_A) ;
    OsSecSubrMake (interp
                   , "kill-process", OsSubrKillProcess
                   , 2, OS_SECURITY_LEVEL_A) ;
    OsSecSubrMake (interp
                   , "with-input-from-process", OsSubrWithInputFromProcess
                   , 2, OS_SECURITY_LEVEL_A);
    OsSecSubrMake (interp
                   , "with-output-to-process"
                   , OsSubrWithOutputToProcess
                   , 2, OS_SECURITY_LEVEL_A) ;
    OsPathProvides (interp, OsStringMake ("process")) ;
    return ;
}    
