;; scheduler.scm
;; scheduler.scm,v 1.2 1995/03/17 14:29:06 baird Exp
;; See the file COPYRIGHT for licensing conditions.
;; AUTODOC MODULE
;; A WWW request scheduler, intended to respect the Ethic for Web spiders.
;; This scheduler should provide an ethical way of making asynchronous requests
;; to httpd servers. 
;; It requires libWWW >= 3.0

(require "www-3.0")
(require "lstutls")

;; Debugging stuff:

(define (debug . rest) #t)

;; Internal API:

;; AUTODOC
;; *www:max-pending-requests*
;; Maximum number of requets that can be launched in parallel.
(define *www:max-pending-request* 3)
;; AUTODOC
;; *www:host-access-delay*
;; Minimum number of seconds to wait between access to the same host.
(define *www:host-access-delay*   2)
;; AUTODOC
;; *www:pending-requests*
;; List of requests waiting for completion.
(define *www:pending-requests*    ())
;; AUTODOC
;; *www:ordered-hosts*
;; List of accessed hosts, sorted in last access time (last host last access).
(define *www:ordered-hosts*       ())
;; AUTODOC
;; *www:remaining-requests*
;; Total count of requests that are either currentlty postponed or pending.
(define *www:remaining-requests*  0)

;; AUTODOC
;; *www:visited-hosts*
;; Hashtable of hosts that have already been visited.
;; Each host name (as a string), is mapped to an object that contains the 
;; following host informations:
;;     - last-accessed: seconds since 1/1/70 that we have accessed the host,
;;     - postponed-requests: list of postponed request for this host,
;;     - host: the host name.

(define *www:visited-hosts*       (make-hashtable))

;; AUTODOC 
;; (scheduler:set-continuation! THUNK)
;; When no more requests are pending, the scheduler will invoke THUNK.
;; This function allows to install a THUNK to be executed when the scheduler
;; is done (i.e: no more requests are pending or postponed).

(define www:scheduler-continuation (lambda () 'done))

(define (www:set-scheduler-continuation! f)
  (set! www:scheduler-continuation f))

(define (www:request-terminates request status)
  (set! *www:remaining-requests* (- *www:remaining-requests* 1))
  ;; Remove it from pending requests
  (set! *www:pending-requests* (removeq! *www:pending-requests* request))
  ;; Update target host last access time:
  (www:update-host-access-time (www:parse-host [[request anchor] address]))
  ;; Now launch a new request, if any:
  (www:request-slot-available))

(define (www:register-pending-request request) 
  (set! *www:remaining-requests* (+ 1 *www:remaining-requests*))
  ;; Hook this request termination method:
  (let ((oldimp [request getImplementationFor: (obj:selector "terminate:")]))
    (define-message request (terminate: status)
      (www:request-terminates self status)
      [self runImplementation: oldimp with: (vector status)]
      (when (= *www:remaining-requests* 0)
            ;; The underlying scheduler is terminated
            (www:scheduler-continuation))) )
  ;; Add it to the list of pending requests:
  (set! *www:pending-requests* (cons request *www:pending-requests*)) )

(define (www:launch-request request)
  (if (< (length *www:pending-requests*) *www:max-pending-request*)
      ;; Ok, launch it now:
      (let ((status (www:load request)))
        (debug "*** Request %0 launched.\n" [[request anchor] address])
        (if (eq? status 'www:would-block)
            (www:register-pending-request request)
            (begin (debug "*** Request %0 terminates right away!\n" request)
                   status)))
      ;; Wait until a slot in *www:pending-requests* is available:
      (www:postpone-request request)))
        

;; Host info management:

(define www:get-host-info 
  (let ((host-info [obj:root clone]))
    [host-info defineSlot: "last-accessed"]
    [host-info defineSlot: "postponed-requests"]
    [host-info defineSlot: "host"]
    ;; Initialize a new host info
    (define-message host-info (new: host)
      (let ((clone [host-info clone])) 
        [clone host: host] 
        [clone last-accessed: 0]
        clone))
    ;; Display host info properly:
    (define-message host-info (display: port)
      (format port "#[HostInfo %0]" [self host]))
    ;; Add a postponed request for this host:
    (define-message host-info (postpone-request: request)
      [self postponed-requests: (cons request [self postponed-requests])])
    ;; Get next postponed request for this host (and remove it from list):
    (define-message host-info (next-postponed-request)
      (let ((postponed [self postponed-requests]))
        [self postponed-requests: (cdr postponed)]
        (car postponed)))
    (lambda (host)
      (or (hashtable-ref *www:visited-hosts* host #f)
          (let ((info [host-info new: host]))
            (hashtable-set! *www:visited-hosts* host info)
            info)))) )

(define (www:postpone-request request)
  (debug "*** Postponing request %0\n" [[request anchor] address])
  ;; (set! *www:remaining-requests* (+ 1 *www:remaining-requests*))
  (let ((info (www:get-host-info (www:parse-host [[request anchor] address]))))
    [info postpone-request: request]
    (or (member [info host] *www:ordered-hosts*)
        (set! *www:ordered-hosts* (cons [info host] *www:ordered-hosts*)))))

(define (www:update-host-access-time host) 
  (let ((host-info (www:get-host-info host)))
    [host-info last-accessed: (current-time)]
    (set! *www:ordered-hosts* 
          (append 
           (list:remove-some! *www:ordered-hosts* 
                              (lambda (i) (string=? host i)))
           (list host)))))

;; This function is called when a request slot in *www:pending-requests* is
;; available. It decides (schedule) what request should be executed next.

(define (www:request-slot-available)
  ;; Look for the first host having a pending request:
  (define (look-for-request)
    (let loop ((host *www:ordered-hosts*))
      (and (pair? host)
           (let ((host-info (www:get-host-info (car host))))
             (if (pair? [host-info postponed-requests])
                 host-info
                 (loop (cdr host)))))))
  (let ((host-info (look-for-request)))
    (when host-info
          (and (< (- (current-time) [host-info last-accessed]) 
                  *www:host-access-delay*)
               (sleep *www:host-access-delay*))
          (www:launch-request [host-info next-postponed-request]))) )
  
;; External API:

;; AUTODOC AUTOLOAD
;; (www:schedule-request REQUEST)
;; Schedule request for execution.
;; The request is registered in a list of pending request for execution. When
;; possible (i.e. compatible with WWW ethic), this request will be started
;; through the 'www:load command.

;; Because this function performs side-effect on the given request, it clones
;; it before, in order not to change what the user expect the behavior to be.
;; The side effect in question here is the hook'ing of the terminate 
;; message implementation of the given request.
;; All this is amidetly too complicated...

(define (www:schedule-request request) 
  [request blocking: #f]
  (www:launch-request [request clone]))

(provide "scheduler")
