;; walker.scm
;; walker.scm,v 1.2 1995/03/17 14:28:07 baird Exp
;; See the file COPYRIGHT for licensing conditions.
;; AUTODOC MODULE
;; An asynchronous WWW walker. This walker uses the ethical scheduler to walk
;; through the WEB, allowiing the programmer to filter what part of it should
;; be walked through.

(require "scheduler")

;; AUTODOC AUTOLOAD
;; (www:walk ANCHOR HOOK FILTER IMAGE-HOOK)
;; Walk through the WEB, starting at ANCHOR.
;; This function uses the scheduler to walk through the WEB starting
;; at ANCHOR.  Each times it encounters a new (child) anchor, it calls
;; FILTER, with the new anchor its HREF attribute (as a string), and
;; its distance from the root ANCHOR. It this returns #t, then the
;; child anchor is visited, otherwise, this branch of the tree is
;; pruned.  
;; For all loaded anchor, the HOOK is called wit hthe anchor and its
;; load status as a parameter.
;; Finally, all IMAGE elements of the source are provided as anchors
;; to the IMAGE-HOOK hook.
;; When using this function, you can set the scheduler continuation to cleanly
;; terminate your script (see 'set-scheduler-continuation)

(define www:walk 
  (let ((request (www:make-request))
        (handler (www:make-html-handler)))
    [handler defineSlot: "depth"      init: 0]
    [handler defineSlot: "filter"     init: (lambda (a h d) #t)]
    [handler defineSlot: "hook"       init: (lambda (a s) anchor)]
    [handler defineSlot: "image-hook" init: (lambda (i s) #f)]
    [handler defineSlot: "request"    init: #f]
    ;; Request termination invokes the user provided hook, which is stored
    ;; in the html handler object.
    (define-message request (terminate: status)
      ([[self handler] hook] [self anchor] status))
    ;; Create a new handler 
    ;; We keep track of the "distance" of the anchor from the root anchor
    (define-message handler (cloneWithDepth: depth)
      (let ((new-handler [self clone]))
        [new-handler depth: depth]
        new-handler) )
    ;; Handle new anchors in the loaded anchor.
    ;; Invoke the user provided filter, if its okay, launch a new request.
    (define-message handler (anchor: anchor href: href)
      (let ((target [anchor mainLink]))
        (when (and target ([self filter] target href [self depth]))
              ;; List this target
              (let ((new-request [[self request] clone])
                    (new-handler [self cloneWithDepth: (+ [self depth] 1)]))
                [new-handler request: new-request]
                [new-request anchor: target]
                [new-request handler: new-handler]
                (www:schedule-request new-request)))) )
    ;; Handling image is easy !
    (define-message handler (image: image src: source)
      ([self image-hook] image source))
    ;; The www:walk closure:
    (lambda (anchor hook filter image-hook)
      (let ((new-request [request clone])
            (new-handler [handler clone]))
        (when (filter anchor [anchor address] 0)
              [new-handler hook: hook]
              [new-handler filter: filter]
              [new-handler request: new-request]
              [new-request anchor: anchor]
              [new-request handler: new-handler]
              (www:schedule-request new-request)))) ) )

(define (test)
  (define filter 
    (let ((visited (make-hashtable)))
      (lambda (anchor depth)
        (when (and (< depth 1)
                   (string=? (www:parse-access [anchor address]) "http")
                   (not (hashtable-ref visited [anchor address] #f)))
              (hashtable-set! visited [anchor address] anchor)
              (format (current-output-port)
                      "Launching request %0 (%1)\n" [anchor address] depth)
              #t))) )
  (define (hook anchor status)
    (format (current-output-port) "%0: %1\n" [anchor address] status))
  (www:walk (www:make-anchor "http://www.inria.fr/koala/abaird.html")
            hook
            filter))

(provide "walker")
