/***************************************************************************
 * w3scheme.c
 * w3scheme.c,v 1.4 1995/03/17 14:35:33 baird Exp
 * See the file COPYRIGHT for licensing conditions.
 * BUGS:
 * - request and stream might be subject to memory leaks.
 * AUTODOC MODULE
 * An interface to the CERN libWWW for oscheme.
 ***************************************************************************/

#include <stdio.h>
#include <setjmp.h>

#include <oscheme.h>

#include "HTUtils.h"	/* WWW general purpose macros */

#include "HTFormat.h"
#include "HTTCP.h"	/* TCP/IP utilities */
#include "HTAnchor.h"   /* Anchor class */
#include "HTParse.h"    /* WWW address manipulation */
#include "HTAccess.h"   /* WWW document access network code */
#include "HTHistory.h"	/* Navigational aids */
#include "HTML.h"	/* For parser */
#include "HTFWriter.h"	/* For non-interactive output */
#include "HTMLGen.h"	/* For reformatting HTML */
#include "HTFile.h"	/* For Dir access flags */
#include "HTRules.h"    /* For loading rule file */
#include "HTError.h"

#include "w3scheme.h"

/*
 * Stuff to be supplied to WWW library :-(
 */

PUBLIC char *HTAppName    = "W3-Scheme" ;
PUBLIC char *HTAppVersion = 0 ;
PUBLIC char *HTBinDir     = (char *) 0 ;

/*
 * The oscheme stream support:
 * ===========================
 */

struct _HTStream {
    CONST HTStreamClass *isa ;		/* generic part */

    OsEObj              os_stream ;	/* scheme stream object */
} ;

/*
 * OScheme stream:
 * Those objects support generic part of lkibWWW stream
 */

static OsEObj stream_class = (OsEObj) 0 ;

static OsEObjSel stream_sel_ptr           = (OsEObjSel) 0 ;
static OsEObjSel stream_sel_put_character = (OsEObjSel) 0 ;
static OsEObjSel stream_sel_put_string    = (OsEObjSel) 0 ;
static OsEObjSel stream_sel_put_block     = (OsEObjSel) 0 ;

static HTStream *HTOScheme_new_with_eobj (OsEObj stream) ;

static OsObject stream_char (closure, self, sel, os_char)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsChar    os_char ;
{
    OsOutputPortWriteChar (OsCurrentOutputPort, os_char->c) ;
    return ((OsObject) os_char) ;
}

static OsObject stream_string (closure, self, sel, os_string)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsString  os_string ;
{
    OsOutputPortWriteCString (OsCurrentOutputPort, os_string->string) ;
    return ((OsObject) os_string) ;
}

static OsObject stream_block (closure, self, sel, os_block)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsString  os_block ;
{
    OsOutputPortWriteCString (OsCurrentOutputPort, os_block->string) ;
    return ((OsObject) os_block) ;
}

static OsObject stream_clone (closure, self, sel)
    void *closure ;
    OsEObj self ;
    OsEObjSel sel ;
{
    OsEObj s = (OsEObj) OsEObjSendMsg0 (stream_class->prototype
                                        , self
                                        , OsEObjSelClone) ;
    HTStream *me = HTOScheme_new_with_eobj (s) ;
    return ((OsObject) me->os_stream) ;
}

static void stream_class_initialize () {
    stream_class = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                            , (OsEObj) OsEObjRoot
                                            , OsEObjSelClone) ;
    OsEObjAllocateSlot (stream_class
                        , stream_sel_ptr = OsEObjSelDefine ("_ptr")
                        , (OsEObjSel) 0
                        , OsA_False) ;
    OsEObjDefineMsg (stream_class
                     , stream_sel_put_character = OsEObjSelDefine ("char:")
                     , stream_char
                     , (void *) 0) ;
    OsEObjDefineMsg (stream_class
                     , stream_sel_put_string = OsEObjSelDefine ("string:")
                     , stream_string
                     , (void *) 0) ;
    OsEObjDefineMsg (stream_class
                     , stream_sel_put_block = OsEObjSelDefine ("block:")
                     , stream_block
                     , (void *) 0) ;
    OsEObjDefineMsg (stream_class
                     , OsEObjSelClone
                     , stream_clone
                     , (void *) 0) ;
    return ;
}

/*
 * LibWWW implementation of stream generic part: just callback the underlying
 * scheme object.
 */

static void HTOScheme_put_character (me, c)
    HTStream *me ;
    char     c ;
{
    OsEObjSendMsg1 (me->os_stream, me->os_stream
                    , stream_sel_put_character
                    , (OsObject) OsCharMake ((unsigned char) c)) ;
    return ;
}

static void HTOScheme_put_string (me, s)
    HTStream *me ;
    char     *s ;
{
    OsEObjSendMsg1 (me->os_stream, me->os_stream
                    , stream_sel_put_string
                    , (OsObject) OsStringMake (s)) ;
    return ;
}

static void HTOScheme_write (me, blk, sz)
    HTStream *me ;
    char     *blk ;
    int      sz ;
{
    OsEObjSendMsg1 (me->os_stream, me->os_stream
                    , stream_sel_put_block
                    , (OsObject) OsStringMakeFromBytes (sz, blk)) ;
    return ;
}

static int HTOScheme_free (HTStream *me) {
    /* fprintf (stderr, "HTOScheme_free x%08x\n", me) ; */ 
    return (0) ;
}

static int HTOScheme_abort (HTStream *me, HTError err) {
    fprintf (stderr, "HTOScheme_abort x%08x\n", me) ;
    return ;
}

HTStreamClass HTOScheme = {
    "OScheme",
    HTOScheme_free,
    HTOScheme_abort,
    HTOScheme_put_character,
    HTOScheme_put_string,
    HTOScheme_write
} ;

static HTStream *HTOScheme_new ()
{
    OsEObj os_stream = (OsEObj) OsEObjSendMsg0 (stream_class, stream_class
                                                , OsEObjSelClone) ;
    OsPointer ptr = (OsPointer) OsEObjSendMsg0 (os_stream, os_stream
                                                , stream_sel_ptr) ;
    return ((HTStream *) ptr->pointer) ;
}

static HTStream *HTOScheme_new_with_eobj (os_stream)
    OsEObj os_stream ;
{
    HTStream *me = (HTStream *) malloc (sizeof (*me)) ;
    if ( ! me )
        outofmem (__FILE__, "HTOScheme new") ;
    me->isa       = &HTOScheme ;
    me->os_stream = os_stream ;
    OsEObjAllocateSlot (me->os_stream
                        , stream_sel_ptr
                        , (OsEObjSel) 0
                        , (OsObject) OsSharedPointerMake (me)) ;
    return ((HTStream *) me) ;
}

/*
 * The oscheme request support:
 * ============================
 */

static OsHashtable request_wrappers = (OsHashtable) 0 ;

static OsEObj request_class = (OsEObj) 0 ;

static OsEObjSel request_sel_handler   = (OsEObjSel) 0 ;
static OsEObjSel request_sel_oformat   = (OsEObjSel) 0 ;
static OsEObjSel request_sel_ostream   = (OsEObjSel) 0 ;
static OsEObjSel request_sel_method    = (OsEObjSel) 0 ;
static OsEObjSel request_sel_anchor    = (OsEObjSel) 0 ;
static OsEObjSel request_sel_terminate = (OsEObjSel) 0 ;
#ifdef WWW_3_0
static OsEObjSel request_sel_blocking = (OsEObjSel) 0 ;
#endif /* WWW_3_0 */

static OsObject request_terminated (closure, self, sel, status)
    void      *closure ;
    OsEObj    self ;
    OsEObjSel sel ;
    OsInteger status ;
{
    /* Nothing done is the default */
    return ((OsObject) status) ;
}

static void request_class_initialize () {
    request_wrappers = OsHashtableMake (0
                                        , (OsHashFct) OsHashPointer
                                        , (OsCmpFct)  OsCmpPointer) ;
    request_class = (OsEObj) OsEObjSendMsg0 ((OsEObj) OsEObjRoot
                                             , (OsEObj) OsEObjRoot
                                             , OsEObjSelClone) ;
    request_sel_handler = OsEObjSelDefine ("handler") ;
    OsEObjAllocateSlot (request_class
                        , request_sel_handler
                        , OsEObjSelDefine ("handler:")
                        , OsA_Unbound) ;
    request_sel_oformat = OsEObjSelDefine ("output-format") ;
    OsEObjAllocateSlot (request_class
                        , request_sel_oformat 
                        , OsEObjSelDefine ("output-format:")
                        , OsA_Unbound) ;
    request_sel_ostream = OsEObjSelDefine ("output-stream") ;
    OsEObjAllocateSlot (request_class
                        , request_sel_ostream
                        , OsEObjSelDefine ("output-stream:") 
                        , OsA_Unbound) ;
    request_sel_method = OsEObjSelDefine ("method") ;
    OsEObjAllocateSlot (request_class
                        , request_sel_method
                        , OsEObjSelDefine ("method:")
                        , OsA_Unbound) ;
    request_sel_anchor = OsEObjSelDefine ("anchor") ;
    OsEObjAllocateSlot (request_class
                        , request_sel_anchor
                        , OsEObjSelDefine ("anchor:")
                        , OsA_Unbound) ;
#ifdef WWW_3_0
    request_sel_blocking = OsEObjSelDefine ("blocking") ;
    OsEObjAllocateSlot (request_class
                        , request_sel_blocking
                        , OsEObjSelDefine ("blocking:")
                        , OsA_Unbound) ;
#endif /* WWW_3_0 */
    request_sel_terminate = OsEObjSelDefine ("terminate:") ;
    OsEObjDefineMsg (request_class
                     , request_sel_terminate
                     , request_terminated
                     , (void *) 0) ;
    return ;
}

/* C API */

HTRequest *OsW3MakeRequest (os_req)
    OsEObj os_req ;
{
    HTRequest *req = HTRequest_new () ;

    /* HTML initialization */
    HTFormatInitNIM (req->conversions) ;
    req->context = (void *) os_req ;
    /* Copy output_format back to request */
    {
        OsAtom ofmt = (OsAtom) OsEObjSendMsg0 (os_req, os_req
                                               , request_sel_oformat) ;
        if ( ofmt != (OsAtom) OsA_Unbound ) {
            OsShouldBeAtom (ofmt) ;
            req->output_format = HTAtom_for (ofmt->pname->string) ;
        }
    }
    /* Copy output stream back to request */
    {
        OsEObj    ostream = (OsEObj) OsEObjSendMsg0 (os_req, os_req
                                                     , request_sel_ostream) ;
        OsPointer ptr ;
        if ( ostream != (OsEObj) OsA_Unbound ) {
            OsShouldBeEObj (ostream) ;
            ptr = (OsPointer) OsEObjSendMsg0 (ostream,ostream,stream_sel_ptr) ;
            OsShouldBePointer (ptr) ;
            req->output_stream = (HTStream *) ptr->pointer ;
        }
    }
    /* Copy method name back to request */
    {
        OsAtom name = (OsAtom) OsEObjSendMsg0 (os_req, os_req
                                               , request_sel_method) ;
        if ( name != (OsAtom) OsA_Unbound ) {
            OsShouldBeAtom (name) ;
            req->method = HTMethod_enum (name->pname->string) ;
        }
    }
    /* Copy anchor back to request */
    {
        OsEObj os_anchor = (OsEObj) OsEObjSendMsg0 (os_req, os_req
                                                    , request_sel_anchor) ;
        if ( os_anchor != (OsEObj) OsA_Unbound ) {
            HTAnchor *anchor ;
            
            OsShouldBeEObj (os_anchor) ;
            anchor = OsW3AnchorOfAnchor (os_anchor) ;
            req->anchor      = HTAnchor_parent (anchor) ;
            req->childAnchor = (((HTAnchor *) req->anchor == anchor)
                                ? (HTChildAnchor *) NULL
                                : (HTChildAnchor *) anchor) ;
        }
    }
#ifdef WWW_3_0
    {
        OsObject os_blocking = OsEObjSendMsg0 (os_req, os_req
                                               , request_sel_blocking) ;
        /* This test ensures default behavior if nothing is specified */
        if ( os_blocking == OsA_False ) { 
            req->BlockingIO = 0 ;
        } else {
            req->BlockingIO = 1 ;
        }
    }
#endif /* WWW_3_0 */
    OsHashtablePut (request_wrappers, req, os_req) ;
    return (req) ;
}

#ifdef WWW_3_0
void OsW3RequestTerminate (req, status)
    HTRequest *req ;
    int       status ;
{
    if ( status == HT_WOULD_BLOCK ) {
        /* Request not terminated yet */
        return ;
    } else {
        OsEObj os_req = (OsEObj) OsHashtableGet (request_wrappers, req) ;

        if ( os_req ) {
            switch (status) {
              case HT_OK:
              case HT_LOADED:
              case HT_NO_DATA:
                (void) OsEObjSendMsg1 (os_req, os_req, request_sel_terminate
                                       , OsA_True) ;
                break ;
              default:
                (void) OsEObjSendMsg1 (os_req, os_req, request_sel_terminate
                                       , OsA_False) ;
                break ;
            }
        }
        OsHashtableRemove (request_wrappers, req) ;
        HTRequest_delete (req) ;
        return ;
    }
    /* NOT REACHED */
}
#endif /* WWW_3_0 */

OsEObj OsW3HandlerOfRequest (req)
    HTRequest *req ;
{
    OsEObj os_req = (OsEObj) req->context ;
    OsEObj hdlr   = (OsEObj) OsEObjSendMsg0 (os_req, os_req
                                             , request_sel_handler) ;
    if ( OsIsEObj (hdlr) )
        return (hdlr) ;
    else
        return ((OsEObj) 0) ;
}

/*
 * Subrs:
 */

/* AUTODOC
 * (www:make-request)
 * Creates an empty WWW request.
 * This function creates an empty WWW request implementing the following set
 * of messages:
 *
 *    (handler): the slot containing the HTML handler for this request. This
 *        handler defines various callbacks to be triggered when
 *        encountering some particular SGML tag in an HTML document (see
 *        'www:make-html-handler). To be activated, the request
 *        output-format should be set to 'www/present, which is the default, 
 *    (output-format): the slot containing the desired output-format
 *        for the reply of the request. This can be any of the documented
 *        output-format for the WWW library. It should be provided as an atom.
 *    (output-stream): the slot containing the output stream to be used for
 *        this request. The usage of this stream depends on the value of the
 *        output-format for the underlying request. To create streams, use the
 *        'www:make-stream function.
 *    (method): the slot holding method to be performed on the anchor of the
 *        request. This can be any method supported by libWWW, and the anchor
 *        (of course). It should be provided as an atom (i.e: 'HEAD) 
 *    (anchor): the slot holding the anchor to be acted upon,
 *    (blocking): a slot, considered as a boolean, indicating whether the
 *        request should block for completion. The termination of non blocking
 *        requets can be hooked by defining a (terminate:) message, which will
 *        be triggered upon termination with the request status,
 *    (terminate:) when a non-blocking request completes, its (terminate:)
 *        message implementation will be triggered with the status as its uniq
 *        parameter.
 *
 * For example, copying the HREF "http://www.inria.fr/koala/abaird,html" into
 * the local file "abaird.html" could be implemented as:
 *
 * ;; Create an HTML stream which dumps its content to a local file:
 * (define (make-copy-stream filename)
 *    (let* ((stream (www:make-stream))
 *           (port   (open-output-file filename)))
 *       (define-message stream (char: c)   (display c port))
 *       (define-message stream (string: s) (display s port))
 *       (define-message stream (block: b)  (display b port))
 *       (define-message stream (close) (close-output-port port))
 *       stream))
 * ;; Create a fresh request:
 * (define request (www:make-request))
 * [request anchor: (www:make-anchor "http://www.inria.fr/koala/abaird.html")]
 * [request output-format: 'text/html]
 * [request output-stream: (make-copy-stream "abaird.html")]
 * ;; Load it
 * (www:load request)
 */

static OsObject OsSubrW3MakeRequest () {
    OsEObj    os_req = (OsEObj) OsEObjSendMsg0 (request_class, request_class
                                                , OsEObjSelClone) ;
    return ((OsObject) os_req) ;
}

/* AUTODOC
 * (www:make-stream)
 * Creates a new WWW stream.
 * This function creates a new WWW streams. This stream defines the following
 * set of messages:
 *    (char:) invoked when a single character is received,
 *    (string:) invoked when a full ASCII string is received,
 *    (block:) invoked when a raw blcok of data is received.
 * The normal usage is to create a WWW stream by using this primitive, and
 * then refines these messages implementation to whatever you want. The defaut
 * implementation dumps all data to standard output.
 *
 * For example, the following function creates a stream that dumps all the
 * data it receives to a local file:
 * (define (make-copy-stream filename)
 *    (let* ((stream (www:make-stream))
 *           (port   (open-output-file filename)))
 *       (define-message stream (char: c)   (display c port))
 *       (define-message stream (string: s) (display s port))
 *       (define-message stream (block: b)  (display b port))
 *       (define-message stream (close) (close-output-port port))
 *       stream))
 */

static OsObject OsSubrW3MakeStream () {
    HTStream *stream   = HTOScheme_new () ;
    return ((OsObject) stream->os_stream) ;
}

/* AUTODOC
 * (www:load REQUEST)
 * Execute the provided World Wide Web REQUEST.
 * Depending on your version of libWWW (3.0 highly recommended), the following
 * result is returned:
 * WWW3.0:
 *    On success, #t is returned, on failure #f. If the request was
 *    specified as non blocking, and the underlying operation would
 *    block, than it returns 'www:would-block.
 * WWW2.16:
 *    On success, #t is returned (the document has been loaded), and
 *    #f on error.
 * See also 'www:make-request.
 */

static OsObject OsSubrW3Load (os_req)
    OsEObj os_req ;
{
    HTRequest *req ;
#ifdef WWW_3_0
    int       status ;
#else
    BOOL      status ;
#endif
    
    OsShouldBeEObj (os_req) ;
    req = OsW3MakeRequest (os_req) ;
#ifdef WWW_3_0    
    status = HTLoad (req, YES) ;
    switch (status) {
      case HT_WOULD_BLOCK:
        return ((OsObject) OsAtomMake (OsCurrentInterp()
                                       , "www:would-block", OsA_Unbound)) ;
      case HT_NO_DATA:
      case HT_LOADED:
        OsW3RequestTerminate (req, status) ;
        return (OsA_True) ;
      case HT_ERROR:
      default:
        OsW3RequestTerminate (req, status) ;
        return (OsA_False) ;
    }
#else /* WWW < 3.x */
    status = HTLoad (req) ;
    return (status ? OsA_True : OsA_False) ;
#endif /* WWW_3_0 */
    /* NOT REACHED */
}

/* AUTODOC
 * (www:parse-host URL)
 * Returns the host part of given URL.
 * This primitive parses URL to find its host. It returns #f if none was found.
 */

static OsObject OsSubrW3ParseHost (OsString name) {
    char     *part ;
    OsShouldBeString (name) ;
    if ( part = HTParse (name->string, name->string, PARSE_HOST) ) {
        OsObject os_ret = (*part ? (OsObject) OsStringMake (part) : OsA_False);
        free (part) ;
        return (os_ret) ;
    }
    return (OsA_False) ;
}

/* AUTODOC
 * (www:parse-access URL)
 * Returns the accesss part of given URL.
 * This primitive parses URL to find its access. It returns #f if none was
 * found.
 */

static OsObject OsSubrW3ParseAccess (OsString name) {
    char     *part ;
    OsShouldBeString (name) ;
    if ( part = HTParse (name->string, name->string, PARSE_ACCESS) ) {
        OsObject os_ret = (*part ? (OsObject) OsStringMake (part) : OsA_False);
        free (part) ;
        return (os_ret) ;
    }
    return (OsA_False) ;
}

/* AUTODOC
 * (www:parse-path URL)
 * Returns the path part of given URL.
 * This primitive parses URL to find its path. It returns #f if none was found,
 * a string otherwise.
 */

static OsObject OsSubrW3ParsePath (OsString name) {
    char *part ;
    OsShouldBeString (name) ;
    if ( part = HTParse (name->string, name->string, PARSE_PATH) ) {
        OsObject os_ret = (*part ? (OsObject) OsStringMake (part) : OsA_False);
        free (part) ;
        return (os_ret) ;
    }
    return (OsA_False) ;
}
    
/*
 * Initialize:
 */

void Os_oswww_initialize (OsInterp interp) {
    extern void Os_anchor_initialize () ;
    extern void Os_HTML_initialize () ;
#ifdef WWW_3_0
    HTLibInit () ;
    Os_w3thread_initialize (interp) ;
#endif /* WWW_3_0 */
    stream_class_initialize () ;
    request_class_initialize () ;
    Os_anchor_initialize (interp) ;
    Os_HTML_initialize (interp) ;
    OsSubrMake (interp, "www:make-stream", OsSubrW3MakeStream, 0) ;
    OsSubrMake (interp, "www:parse-host", OsSubrW3ParseHost, 1) ;
    OsSubrMake (interp, "www:parse-access", OsSubrW3ParseAccess, 1) ;
    OsSubrMake (interp, "www:parse-path", OsSubrW3ParsePath, 1) ;
    OsSubrMake (interp, "www:make-request", OsSubrW3MakeRequest, 0) ;
    OsSubrMake (interp, "www:load", OsSubrW3Load, 1) ;
    OsPathProvides (interp, OsStringMake ("www")) ;
    return ;
}
